"""Benchmark type checker performance with and without generated stubs.

Measures wall-clock time for pyright, mypy, and ty to type-check:
1. The full idfkit package (src/idfkit/)
2. A small consumer script that exercises typed APIs

Run:
    uv run python benchmarks/bench_typecheckers.py
"""

from __future__ import annotations

import shutil
import subprocess
import sys
import tempfile
import textwrap
import time
from pathlib import Path

ROOT = Path(__file__).resolve().parent.parent
SRC = ROOT / "src" / "idfkit"
STUBS = [SRC / "_generated_types.pyi", SRC / "document.pyi"]

# A small script that exercises the typed API surface
CONSUMER_SCRIPT = textwrap.dedent("""\
    from idfkit import new_document, load_idf, IDFDocument

    doc = new_document()
    zones = doc["Zone"]
    zone = doc.add("Zone", "Office")
    name: str = zone.name

    strict_doc = new_document(strict=True)
    copy = strict_doc.copy()
""")

RUNS = 3  # number of runs per benchmark


def _run_timed(cmd: list[str], label: str, *, cwd: Path | None = None) -> float:
    """Run a command, return wall-clock seconds. Returns -1 on failure."""
    times: list[float] = []
    for i in range(RUNS):
        start = time.perf_counter()
        result = subprocess.run(cmd, capture_output=True, text=True, cwd=cwd or ROOT)
        elapsed = time.perf_counter() - start
        times.append(elapsed)
        if i == 0 and result.returncode != 0:
            errors = [line for line in result.stdout.splitlines() if "error" in line.lower()]
            if errors:
                print(f"    {label}: {len(errors)} error(s) on first run")

    avg = sum(times) / len(times)
    return avg


def _hide_stubs() -> list[tuple[Path, Path]]:
    """Temporarily rename stub files so type checkers don't see them."""
    moved: list[tuple[Path, Path]] = []
    for stub in STUBS:
        if stub.exists():
            hidden = stub.with_suffix(".pyi.hidden")
            stub.rename(hidden)
            moved.append((hidden, stub))
    return moved


def _restore_stubs(moved: list[tuple[Path, Path]]) -> None:
    """Restore hidden stub files."""
    for hidden, original in moved:
        if hidden.exists():
            hidden.rename(original)


def bench_pyright_package(with_stubs: bool) -> float:
    """Benchmark pyright on the full package."""
    cmd = [sys.executable, "-m", "pyright", str(SRC)]
    label = f"pyright package ({'with' if with_stubs else 'without'} stubs)"
    return _run_timed(cmd, label)


def bench_pyright_consumer(consumer_path: Path, with_stubs: bool) -> float:
    """Benchmark pyright on a small consumer script."""
    cmd = [sys.executable, "-m", "pyright", str(consumer_path)]
    label = f"pyright consumer ({'with' if with_stubs else 'without'} stubs)"
    return _run_timed(cmd, label)


def bench_mypy_package(with_stubs: bool) -> float:
    """Benchmark mypy on the full package."""
    cmd = [
        sys.executable,
        "-m",
        "mypy",
        "--no-incremental",
        "--ignore-missing-imports",
        str(SRC),
    ]
    label = f"mypy package ({'with' if with_stubs else 'without'} stubs)"
    return _run_timed(cmd, label)


def bench_mypy_consumer(consumer_path: Path, with_stubs: bool) -> float:
    """Benchmark mypy on a small consumer script."""
    cmd = [
        sys.executable,
        "-m",
        "mypy",
        "--no-incremental",
        "--ignore-missing-imports",
        str(consumer_path),
    ]
    label = f"mypy consumer ({'with' if with_stubs else 'without'} stubs)"
    return _run_timed(cmd, label)


def bench_ty_package(with_stubs: bool) -> float:
    """Benchmark ty on the full package."""
    ty_bin = shutil.which("ty")
    if not ty_bin:
        return -1.0
    cmd = [ty_bin, "check", str(SRC)]
    label = f"ty package ({'with' if with_stubs else 'without'} stubs)"
    return _run_timed(cmd, label)


def bench_ty_consumer(consumer_path: Path, with_stubs: bool) -> float:
    """Benchmark ty on a small consumer script."""
    ty_bin = shutil.which("ty")
    if not ty_bin:
        return -1.0
    cmd = [ty_bin, "check", str(consumer_path)]
    label = f"ty consumer ({'with' if with_stubs else 'without'} stubs)"
    return _run_timed(cmd, label)


def _format_time(seconds: float) -> str:
    if seconds < 0:
        return "N/A"
    return f"{seconds:.2f}s"


def _format_delta(with_val: float, without_val: float) -> str:
    if with_val < 0 or without_val < 0:
        return "N/A"
    delta = with_val - without_val
    pct = (delta / without_val) * 100 if without_val > 0 else 0
    sign = "+" if delta > 0 else ""
    return f"{sign}{delta:.2f}s ({sign}{pct:.0f}%)"


def main() -> None:
    print("=" * 72)
    print("Type Checker Benchmark: idfkit with vs without generated stubs")
    print("  Stubs: _generated_types.pyi + document.pyi")
    print("  Stub sizes: ", end="")
    for s in STUBS:
        if s.exists():
            print(f"{s.name}={s.stat().st_size / 1024:.0f}KB  ", end="")
    print()
    print(f"  Runs per benchmark: {RUNS} (reporting average)")
    print("=" * 72)

    # Write consumer script to a temp file
    tmp_dir = Path(tempfile.mkdtemp(prefix="idfkit_bench_"))
    consumer_path = tmp_dir / "consumer.py"
    consumer_path.write_text(CONSUMER_SCRIPT)

    results: dict[str, dict[str, float]] = {}

    # ---- WITH stubs ----
    print("\n--- WITH stubs ---")
    results["with"] = {}

    print("  Running pyright (package)...")
    results["with"]["pyright_pkg"] = bench_pyright_package(True)

    print("  Running pyright (consumer)...")
    results["with"]["pyright_con"] = bench_pyright_consumer(consumer_path, True)

    print("  Running mypy (package)...")
    results["with"]["mypy_pkg"] = bench_mypy_package(True)

    print("  Running mypy (consumer)...")
    results["with"]["mypy_con"] = bench_mypy_consumer(consumer_path, True)

    print("  Running ty (package)...")
    results["with"]["ty_pkg"] = bench_ty_package(True)

    print("  Running ty (consumer)...")
    results["with"]["ty_con"] = bench_ty_consumer(consumer_path, True)

    # ---- WITHOUT stubs ----
    print("\n--- WITHOUT stubs ---")
    moved = _hide_stubs()
    results["without"] = {}

    try:
        print("  Running pyright (package)...")
        results["without"]["pyright_pkg"] = bench_pyright_package(False)

        print("  Running pyright (consumer)...")
        results["without"]["pyright_con"] = bench_pyright_consumer(consumer_path, False)

        print("  Running mypy (package)...")
        results["without"]["mypy_pkg"] = bench_mypy_package(False)

        print("  Running mypy (consumer)...")
        results["without"]["mypy_con"] = bench_mypy_consumer(consumer_path, False)

        print("  Running ty (package)...")
        results["without"]["ty_pkg"] = bench_ty_package(False)

        print("  Running ty (consumer)...")
        results["without"]["ty_con"] = bench_ty_consumer(consumer_path, False)
    finally:
        _restore_stubs(moved)

    # ---- Results table ----
    print("\n" + "=" * 72)
    print(f"RESULTS (average of {RUNS} runs)")
    print("=" * 72)
    print(f"{'Benchmark':<28} {'Without':>10} {'With':>10} {'Delta':>18}")
    print("-" * 72)

    labels = {
        "pyright_pkg": "pyright (full package)",
        "pyright_con": "pyright (consumer script)",
        "mypy_pkg": "mypy (full package)",
        "mypy_con": "mypy (consumer script)",
        "ty_pkg": "ty (full package)",
        "ty_con": "ty (consumer script)",
    }

    for key, label in labels.items():
        w = results["with"].get(key, -1)
        wo = results["without"].get(key, -1)
        delta = _format_delta(w, wo)
        print(f"  {label:<26} {_format_time(wo):>10} {_format_time(w):>10} {delta:>18}")

    print("-" * 72)

    # Stub file info
    print("\nStub file sizes:")
    for s in STUBS:
        if s.exists():
            size_kb = s.stat().st_size / 1024
            lines = len(s.read_text().splitlines())
            print(f"  {s.name}: {size_kb:.0f} KB, {lines:,} lines")

    # Cleanup
    shutil.rmtree(tmp_dir, ignore_errors=True)


if __name__ == "__main__":
    main()



"""Weather station data model and search result types."""

from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class WeatherStation:
    """Metadata for a single weather file entry from climate.onebuilding.org.

    Each instance represents one downloadable weather dataset. The same physical
    station may appear multiple times with different ``source`` or year-range
    variants (e.g. ``TMYx.2007-2021`` vs ``TMYx.2009-2023``).

    Attributes:
        country: ISO 3166 country code (e.g. ``"USA"``).
        state: State or province abbreviation (e.g. ``"CA"``).
        city: City or station name as it appears in the index
            (e.g. ``"Marina.Muni.AP"``).
        wmo: WMO station number as a string to preserve leading zeros
            (e.g. ``"722950"`` or ``"012345"``).
        source: Dataset source identifier (e.g. ``"SRC-TMYx"``).
        latitude: Decimal degrees, north positive.
        longitude: Decimal degrees, east positive.
        timezone: Hours offset from GMT (e.g. ``-8.0``).
        elevation: Meters above sea level.
        url: Full download URL for the ZIP archive.
    """

    country: str
    state: str
    city: str
    wmo: str
    source: str
    latitude: float
    longitude: float
    timezone: float
    elevation: float
    url: str

    def to_dict(self) -> dict[str, str | float]:
        """Serialize to a plain dictionary for JSON storage."""
        return {
            "country": self.country,
            "state": self.state,
            "city": self.city,
            "wmo": self.wmo,
            "source": self.source,
            "latitude": self.latitude,
            "longitude": self.longitude,
            "timezone": self.timezone,
            "elevation": self.elevation,
            "url": self.url,
        }

    @classmethod
    def from_dict(cls, data: dict[str, str | float]) -> WeatherStation:
        """Deserialize from a plain dictionary."""
        return cls(
            country=str(data["country"]),
            state=str(data["state"]),
            city=str(data["city"]),
            wmo=str(data["wmo"]),
            source=str(data["source"]),
            latitude=float(data["latitude"]),
            longitude=float(data["longitude"]),
            timezone=float(data["timezone"]),
            elevation=float(data["elevation"]),
            url=str(data["url"]),
        )

    @property
    def display_name(self) -> str:
        """Human-readable station name with location context.

        Dots in the city name are replaced with spaces for readability.
        """
        name = self.city.replace(".", " ").replace("-", " ").strip()
        parts: list[str] = []
        if name:
            parts.append(name)
        if self.state:
            parts.append(self.state)
        parts.append(self.country)
        return ", ".join(parts)

    @property
    def filename_stem(self) -> str:
        """The canonical EPW filename stem derived from the download URL.

        Returns the ZIP filename without the ``.zip`` extension, e.g.
        ``"USA_IL_Chicago.Ohare.Intl.AP.725300_TMYx.2009-2023"``.
        """
        filename = self.url.rsplit("/", maxsplit=1)[-1]
        return filename.removesuffix(".zip")

    @property
    def dataset_variant(self) -> str:
        """Extract the TMYx dataset variant from the download URL.

        Returns a string like ``"TMYx"``, ``"TMYx.2007-2021"``, or
        ``"TMYx.2009-2023"``.
        """
        # Dataset variant is everything after the last underscore
        # e.g. "USA_CA_Marina.Muni.AP.690070_TMYx" -> "TMYx"
        # e.g. "USA_CA_Twentynine.Palms.SELF.690150_TMYx.2004-2018" -> "TMYx.2004-2018"
        parts = self.filename_stem.rsplit("_", maxsplit=1)
        if len(parts) == 2:
            return parts[1]
        return self.filename_stem


@dataclass(frozen=True)
class SearchResult:
    """A text search result with relevance score."""

    station: WeatherStation
    score: float
    """Relevance score from 0.0 to 1.0, higher is better."""
    match_field: str
    """Which field matched: ``"wmo"``, ``"name"``, ``"state"``, ``"country"``, or ``"filename"``."""


@dataclass(frozen=True)
class SpatialResult:
    """A spatial proximity result with great-circle distance."""

    station: WeatherStation
    distance_km: float
    """Great-circle distance in kilometres."""



from __future__ import annotations

from idfkit.simulation import SimulationResult, TimeSeriesResult

result: SimulationResult = ...  # type: ignore[assignment]
ts: TimeSeriesResult = ...  # type: ignore[assignment]
# --8<-- [start:example]
ts = result.sql.get_timeseries(
    "Zone Mean Air Temperature",
    "ZONE 1",
)

# Default plot
fig = ts.plot()

# Custom title
fig = ts.plot(title="My Custom Title")

# Explicit backend
from idfkit.simulation import MatplotlibBackend

fig = ts.plot(backend=MatplotlibBackend())
# --8<-- [end:example]



from __future__ import annotations

from idfkit.simulation import SimulationResult
from idfkit.weather import StationIndex, WeatherStation

index: StationIndex = ...  # type: ignore[assignment]
results: list[SimulationResult | None] = ...  # type: ignore[assignment]
station: WeatherStation = ...  # type: ignore[assignment]
# --8<-- [start:example]
from idfkit.weather import StationIndex, geocode

index = StationIndex.load()

# One-liner using splat operator
results = index.nearest(*geocode("Willis Tower, Chicago, IL"))

# First result is the nearest station
station = results[0].station
print(f"Nearest: {station.display_name} ({results[0].distance_km:.1f} km)")
# --8<-- [end:example]



from __future__ import annotations

from idfkit.simulation import S3FileSystem, SimulationResult

fs: S3FileSystem = ...  # type: ignore[assignment]
result: SimulationResult = ...  # type: ignore[assignment]
# --8<-- [start:example]
from idfkit.simulation import SimulationResult

# From a local directory
result = SimulationResult.from_directory("/path/to/sim_output")

# From a cloud storage location
from idfkit.simulation import S3FileSystem

fs = S3FileSystem(bucket="my-bucket")
result = SimulationResult.from_directory("runs/run-001", fs=fs)

# Query data
ts = result.sql.get_timeseries("Zone Mean Air Temperature", "ZONE 1")
# --8<-- [end:example]



from __future__ import annotations

from idfkit.simulation import SimulationProgress
from typing import Any

message_queue: Any = ...  # type: ignore[assignment]
# --8<-- [start:example]
from dataclasses import asdict
import json


def on_progress(event: SimulationProgress) -> None:
    message_queue.publish(json.dumps(asdict(event)))


# --8<-- [end:example]



from __future__ import annotations

# --8<-- [start:example]
import logging

# Turn on DEBUG for the simulation subsystem
logging.basicConfig(
    level=logging.WARNING,
    format="%(asctime)s %(name)s %(levelname)s %(message)s",
)
logging.getLogger("idfkit.simulation").setLevel(logging.DEBUG)

# Now run a simulation — every step is visible:
#   DEBUG idfkit.simulation.config  Trying candidate /usr/local/EnergyPlus-24-2-0
#   INFO  idfkit.simulation.config  Found EnergyPlus 24.2.0 at /usr/local/EnergyPlus-24-2-0
#   DEBUG idfkit.simulation.runner  EnergyPlus: /usr/local/.../energyplus (version 24.2.0)
#   INFO  idfkit.simulation.runner  Starting simulation with weather weather.epw
#   DEBUG idfkit.simulation.runner  Cache miss for key a1b2c3d4e5f6
#   DEBUG idfkit.simulation.runner  Command: /usr/local/.../energyplus -w weather.epw -d ...
#   INFO  idfkit.simulation.runner  Simulation completed successfully in 12.3s
# --8<-- [end:example]



from __future__ import annotations

from idfkit.simulation import SQLResult, SimulationResult

result: SimulationResult = ...  # type: ignore[assignment]
sql: SQLResult | None = ...  # type: ignore[assignment]
# --8<-- [start:example]
sql = result.sql
if sql is None:
    print("No SQL output - was Output:SQLite in the model?")
    return

# Get time series (raises KeyError if not found)
try:
    ts = sql.get_timeseries("Nonexistent Variable", "ZONE 1")
except KeyError as e:
    print(f"Variable not found: {e}")
# --8<-- [end:example]



from __future__ import annotations

from idfkit.simulation import SimulationResult, TimeSeriesResult

result: SimulationResult = ...  # type: ignore[assignment]
ts: TimeSeriesResult = ...  # type: ignore[assignment]
# --8<-- [start:example]
# Get time-series data from SQLite output
ts = result.sql.get_timeseries(
    variable_name="Zone Mean Air Temperature",
    key_value="Office",
)
print(f"Variable: {ts.variable_name}")
print(f"Temperature range: {min(ts.values):.1f}°C to {max(ts.values):.1f}°C")

# Filter by environment if needed
ts_sizing = result.sql.get_timeseries(
    variable_name="Zone Mean Air Temperature",
    key_value="Office",
    environment="sizing",  # Design day data only
)

# Get tabular data
tables = result.sql.get_tabular_data(report_name="AnnualBuildingUtilityPerformanceSummary")
# --8<-- [end:example]



from __future__ import annotations

from datetime import datetime
from idfkit import IDFDocument, IDFObject

doc: IDFDocument = ...  # type: ignore[assignment]
schedule: IDFObject = ...  # type: ignore[assignment]
# --8<-- [start:example]
from idfkit.schedules import evaluate

# Summer design day (typically peak cooling)
value = evaluate(
    schedule,
    datetime(2024, 7, 15, 14, 0),
    day_type="summer",
)

# Winter design day (typically peak heating)
value = evaluate(
    schedule,
    datetime(2024, 1, 15, 6, 0),
    day_type="winter",
)
# --8<-- [end:example]



from __future__ import annotations

from idfkit.simulation import SimulationJob, simulate_batch

jobs: list[SimulationJob] = ...  # type: ignore[assignment]
# --8<-- [start:example]
from rich.progress import Progress

with Progress() as progress:
    task = progress.add_task("Simulating...", total=len(jobs))

    def callback(completed, total, label, success):
        progress.update(task, completed=completed)

    batch = simulate_batch(jobs, progress=callback)
# --8<-- [end:example]



from __future__ import annotations

# --8<-- [start:example]
from scythe.worker import ScytheWorkerConfig

# Import your experiments so they are registered
from experiments.building_energy import simulate_building  # noqa: F401  # pyright: ignore[reportUnusedImport]

if __name__ == "__main__":
    worker_config = ScytheWorkerConfig()
    worker_config.start()
# --8<-- [end:example]



from __future__ import annotations

from idfkit.simulation import SimulationResult

result: SimulationResult = ...  # type: ignore[assignment]
# --8<-- [start:example]
# Check stderr
print(result.stderr)

# Check the run directory
print(f"Outputs in: {result.run_dir}")
# --8<-- [end:example]



from __future__ import annotations

from idfkit import IDFDocument

model: IDFDocument = ...  # type: ignore[assignment]
# --8<-- [start:example]
from idfkit import new_document

# Create a baseline model for EnergyPlus 24.1
model = new_document(version=(24, 1, 0))

# Update pre-seeded singleton objects
building = model["Building"].first()
if building is not None:
    building.name = "My Building"
    building.north_axis = 0
    building.terrain = "City"

geometry_rules = model["GlobalGeometryRules"].first()
if geometry_rules is not None:
    geometry_rules.starting_vertex_position = "UpperLeftCorner"
    geometry_rules.vertex_entry_direction = "Counterclockwise"
    geometry_rules.coordinate_system = "Relative"

# Add additional named objects
model.add("Zone", "Office", x_origin=0, y_origin=0, z_origin=0)

# Add an additional unnamed singleton object
model.add("Timestep", number_of_timesteps_per_hour=4)
# --8<-- [end:example]



from __future__ import annotations

from idfkit.simulation import SQLResult

sql: SQLResult | None = ...  # type: ignore[assignment]
# --8<-- [start:example]
rows = sql.get_tabular_data(
    report_name="AnnualBuildingUtilityPerformanceSummary",
    table_name="Site and Source Energy",
)
# --8<-- [end:example]



from __future__ import annotations

from idfkit import IDFDocument

model: IDFDocument = ...  # type: ignore[assignment]
# --8<-- [start:example]
from idfkit import load_idf
from idfkit.simulation import simulate

model = load_idf("building.idf")
result = simulate(model, "weather.epw", annual=True, on_progress="tqdm")
# --8<-- [end:example]



from __future__ import annotations

from idfkit.weather import WeatherDownloader

# --8<-- [start:example]
from pathlib import Path

downloader = WeatherDownloader(cache_dir=Path("/data/weather_cache"))
# --8<-- [end:example]



from __future__ import annotations

from idfkit import IDFDocument
from idfkit.weather import DesignDayManager

ddm: DesignDayManager = ...  # type: ignore[assignment]
model: IDFDocument = ...  # type: ignore[assignment]
# --8<-- [start:example]
added = ddm.apply_to_model(
    model,
    heating="99%",  # Less extreme
    cooling="1%",
)
# --8<-- [end:example]



from __future__ import annotations

from idfkit.simulation import ErrorReport, SimulationResult

errors: ErrorReport = ...  # type: ignore[assignment]
result: SimulationResult = ...  # type: ignore[assignment]
# --8<-- [start:example]
if not result.success:
    print(errors.summary())
    for err in errors.fatal + errors.severe:
        print(err.message)
# --8<-- [end:example]



from __future__ import annotations

from idfkit import IDFDocument, IDFObject
from idfkit.weather import WeatherStation

added: list[IDFObject] = ...  # type: ignore[assignment]
model: IDFDocument = ...  # type: ignore[assignment]
station: WeatherStation = ...  # type: ignore[assignment]
# --8<-- [start:example]
from idfkit.weather import apply_ashrae_sizing

# Apply ASHRAE 90.1 design conditions
added = apply_ashrae_sizing(model, station, standard="90.1")
print(f"Added {len(added)} design days")
# --8<-- [end:example]



from __future__ import annotations

from idfkit import IDFDocument
from idfkit.simulation import S3FileSystem, simulate
from idfkit.weather import StationIndex, WeatherDownloader, WeatherFiles, WeatherStation

downloader: WeatherDownloader = ...  # type: ignore[assignment]
files: WeatherFiles = ...  # type: ignore[assignment]
fs: S3FileSystem = ...  # type: ignore[assignment]
index: StationIndex = ...  # type: ignore[assignment]
model: IDFDocument = ...  # type: ignore[assignment]
station: WeatherStation = ...  # type: ignore[assignment]
# --8<-- [start:example]
from idfkit.weather import StationIndex, WeatherDownloader

# Download weather file locally
index = StationIndex.load()
station = index.search("chicago")[0].station
downloader = WeatherDownloader()
files = downloader.download(station)

# Then use local weather with S3 output
fs = S3FileSystem(bucket="results", prefix="study/")
result = simulate(
    model,
    files.epw,  # Local path
    output_dir="run-001",
    fs=fs,
)
# --8<-- [end:example]



from __future__ import annotations

from idfkit import IDFDocument

model: IDFDocument = ...  # type: ignore[assignment]
# --8<-- [start:example]
from idfkit.simulation import simulate, SimulationProgress


def on_progress(event: SimulationProgress) -> None:
    match event.phase:
        case "warmup":
            print(f"  Warmup iteration {event.warmup_day}")
        case "simulating":
            pct = f"{event.percent:.0f}%" if event.percent else "?"
            print(f"  [{pct}] Simulating {event.environment}")
        case "complete":
            print("  Simulation complete!")


result = simulate(model, "weather.epw", on_progress=on_progress)
# --8<-- [end:example]



from __future__ import annotations

from idfkit import IDFDocument
from idfkit.simulation import SimulationJob

jobs: list[SimulationJob] = ...  # type: ignore[assignment]
variant: IDFDocument = ...  # type: ignore[assignment]
variants: list[IDFDocument] = ...  # type: ignore[assignment]
# --8<-- [start:example]
import asyncio
from idfkit.simulation import async_simulate_batch_stream, SimulationJob


async def main():
    jobs = [
        SimulationJob(model=variant, weather="weather.epw", label=f"case-{i}") for i, variant in enumerate(variants)
    ]

    async for event in async_simulate_batch_stream(jobs, max_concurrent=4):
        status = "OK" if event.result.success else "FAIL"
        print(f"[{event.completed}/{event.total}] {event.label}: {status}")


asyncio.run(main())
# --8<-- [end:example]



"""Async batch simulation with semaphore-based concurrency control.

Non-blocking counterpart to [simulate_batch][idfkit.simulation.batch.simulate_batch]
that uses [asyncio][] for concurrency instead of
[ThreadPoolExecutor][concurrent.futures.ThreadPoolExecutor].

Two entry points are provided:

* [async_simulate_batch][idfkit.simulation.async_batch.async_simulate_batch] — runs all jobs and returns an aggregated
  [BatchResult][idfkit.simulation.batch.BatchResult], mirroring the sync API.
* [async_simulate_batch_stream][idfkit.simulation.async_batch.async_simulate_batch_stream] — an async generator that yields
  [SimulationEvent][idfkit.simulation.async_batch.SimulationEvent] objects as each simulation completes, enabling
  real-time progress monitoring without callbacks.

Examples:
    ```python
    import asyncio
    from idfkit.simulation import async_simulate_batch, SimulationJob

    async def main():
        jobs = [SimulationJob(model=m, weather="weather.epw", label=f"run-{i}") for i, m in enumerate(models)]
        batch = await async_simulate_batch(jobs, max_concurrent=4)
        print(f"{len(batch.succeeded)}/{len(batch)} succeeded")

    asyncio.run(main())
    ```

Streaming example:

    ```python
    from idfkit.simulation import async_simulate_batch_stream

    async def main():
        jobs = [...]
        async for event in async_simulate_batch_stream(jobs, max_concurrent=4):
            status = "OK" if event.result.success else "FAIL"
            print(f"[{event.completed}/{event.total}] {event.label}: {status}")
    ```
"""

from __future__ import annotations

import asyncio
import logging
import os
import tempfile
import time
from collections.abc import AsyncIterator, Callable, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any

from .async_runner import async_simulate
from .batch import BatchResult, SimulationJob
from .progress import SimulationProgress
from .progress_bars import resolve_on_progress
from .result import SimulationResult

if TYPE_CHECKING:
    from .cache import SimulationCache
    from .config import EnergyPlusConfig
    from .fs import AsyncFileSystem, FileSystem

logger = logging.getLogger(__name__)


@dataclass(frozen=True, slots=True)
class SimulationEvent:
    """Progress event emitted by [async_simulate_batch_stream][idfkit.simulation.async_batch.async_simulate_batch_stream].

    Each event represents a single simulation that has finished (successfully
    or not).  Events are yielded in *completion order*, not submission order.

    Attributes:
        index: Zero-based position of this job in the original *jobs* sequence.
        label: Human-readable label from the [SimulationJob][idfkit.simulation.batch.SimulationJob].
        result: The simulation result.
        completed: Number of jobs completed so far (including this one).
        total: Total number of jobs in the batch.
    """

    index: int
    label: str
    result: SimulationResult
    completed: int
    total: int


async def async_simulate_batch(
    jobs: Sequence[SimulationJob],
    *,
    energyplus: EnergyPlusConfig | None = None,
    max_concurrent: int | None = None,
    cache: SimulationCache | None = None,
    fs: FileSystem | AsyncFileSystem | None = None,
    on_progress: Callable[[SimulationProgress], Any] | None = None,
) -> BatchResult:
    """Run multiple EnergyPlus simulations concurrently using asyncio.

    This is the async counterpart to
    [simulate_batch][idfkit.simulation.batch.simulate_batch].  Concurrency is
    controlled with an [asyncio.Semaphore][] instead of a thread pool.

    Individual job failures are captured as failed
    [SimulationResult][idfkit.simulation.result.SimulationResult] entries -- the batch
    never raises due to a single job failing.

    Args:
        jobs: Sequence of simulation jobs to execute.
        energyplus: Shared EnergyPlus configuration (auto-discovered if
            ``None``).
        max_concurrent: Maximum number of concurrent simulations.  Defaults
            to ``min(len(jobs), os.cpu_count() or 1)``.
        cache: Optional simulation cache for content-hash lookups.
        fs: Optional file system backend passed through to each
            [async_simulate][idfkit.simulation.async_runner.async_simulate] call.
        on_progress: Optional callback invoked with
            [SimulationProgress][idfkit.simulation.progress.SimulationProgress] events
            during each individual simulation.  Events include
            ``job_index`` and ``job_label`` to identify which batch job
            they belong to.  Both sync and async callables are accepted.
            The ``"tqdm"`` shorthand is not supported for batch runners;
            use [tqdm_progress][idfkit.simulation.progress_bars.tqdm_progress]
            with a custom per-job callback instead.

    Returns:
        A [BatchResult][idfkit.simulation.batch.BatchResult] with results in the
        same order as *jobs*.

    Raises:
        ValueError: If *jobs* is empty.
    """
    if not jobs:
        msg = "jobs must not be empty"
        raise ValueError(msg)

    if on_progress == "tqdm":
        msg = (
            'on_progress="tqdm" is not supported for batch simulations because a single '
            "progress bar cannot represent multiple concurrent jobs. Use the tqdm_progress() "
            "context manager with a custom callback instead."
        )
        raise ValueError(msg)

    progress_cb, progress_cleanup = resolve_on_progress(on_progress)

    try:
        if max_concurrent is None:
            max_concurrent = min(len(jobs), os.cpu_count() or 1)

        logger.info("Starting async batch of %d jobs with concurrency %d", len(jobs), max_concurrent)

        semaphore = asyncio.Semaphore(max_concurrent)
        results: list[SimulationResult | None] = [None] * len(jobs)
        start = time.monotonic()

        async def _run_one(idx: int, job: SimulationJob) -> None:
            async with semaphore:
                results[idx] = await _async_run_job(idx, job, energyplus, cache, fs, progress_cb)

        tasks = [asyncio.create_task(_run_one(i, job)) for i, job in enumerate(jobs)]
        await asyncio.gather(*tasks)
    finally:
        if progress_cleanup is not None:
            progress_cleanup()

    elapsed = time.monotonic() - start

    final: list[SimulationResult] = []
    for r in results:
        assert r is not None  # noqa: S101
        final.append(r)

    batch_result = BatchResult(results=tuple(final), total_runtime_seconds=elapsed)
    logger.info(
        "Async batch complete: %d succeeded, %d failed in %.1fs",
        len(batch_result.succeeded),
        len(batch_result.failed),
        elapsed,
    )
    return batch_result


async def async_simulate_batch_stream(
    jobs: Sequence[SimulationJob],
    *,
    energyplus: EnergyPlusConfig | None = None,
    max_concurrent: int | None = None,
    cache: SimulationCache | None = None,
    fs: FileSystem | AsyncFileSystem | None = None,
    on_progress: Callable[[SimulationProgress], Any] | None = None,
) -> AsyncIterator[SimulationEvent]:
    """Run simulations concurrently, yielding events as each one completes.

    This is an async generator variant of [async_simulate_batch][idfkit.simulation.async_batch.async_simulate_batch] that
    yields [SimulationEvent][idfkit.simulation.async_batch.SimulationEvent] objects in *completion order*.  This
    enables real-time progress reporting without needing a callback:

    .. code-block:: python

        async for event in async_simulate_batch_stream(jobs, max_concurrent=4):
            print(f"[{event.completed}/{event.total}] {event.label}")

    Args:
        jobs: Sequence of simulation jobs to execute.
        energyplus: Shared EnergyPlus configuration (auto-discovered if
            ``None``).
        max_concurrent: Maximum number of concurrent simulations.  Defaults
            to ``min(len(jobs), os.cpu_count() or 1)``.
        cache: Optional simulation cache for content-hash lookups.
        fs: Optional file system backend.
        on_progress: Optional callback invoked with
            [SimulationProgress][idfkit.simulation.progress.SimulationProgress] events
            during each individual simulation.  Events include
            ``job_index`` and ``job_label``.  The ``"tqdm"`` shorthand
            is not supported for batch runners; use
            [tqdm_progress][idfkit.simulation.progress_bars.tqdm_progress]
            with a custom per-job callback instead.

    Yields:
        [SimulationEvent][idfkit.simulation.async_batch.SimulationEvent] for each completed simulation, in the order
        they finish.

    Raises:
        ValueError: If *jobs* is empty.
    """
    if not jobs:
        msg = "jobs must not be empty"
        raise ValueError(msg)

    if on_progress == "tqdm":
        msg = (
            'on_progress="tqdm" is not supported for batch simulations because a single '
            "progress bar cannot represent multiple concurrent jobs. Use the tqdm_progress() "
            "context manager with a custom callback instead."
        )
        raise ValueError(msg)

    progress_cb, progress_cleanup = resolve_on_progress(on_progress)

    if max_concurrent is None:
        max_concurrent = min(len(jobs), os.cpu_count() or 1)

    semaphore = asyncio.Semaphore(max_concurrent)
    total = len(jobs)
    queue: asyncio.Queue[SimulationEvent] = asyncio.Queue()
    completed_count = 0

    async def _run_one(idx: int, job: SimulationJob) -> None:
        nonlocal completed_count
        async with semaphore:
            result = await _async_run_job(idx, job, energyplus, cache, fs, progress_cb)
        completed_count += 1
        await queue.put(
            SimulationEvent(
                index=idx,
                label=job.label,
                result=result,
                completed=completed_count,
                total=total,
            )
        )

    tasks = [asyncio.create_task(_run_one(i, job)) for i, job in enumerate(jobs)]

    try:
        for _ in range(total):
            event = await queue.get()
            yield event
    finally:
        if progress_cleanup is not None:
            progress_cleanup()
        # If the consumer breaks out early, cancel remaining tasks.
        for task in tasks:
            task.cancel()
        await asyncio.gather(*tasks, return_exceptions=True)


async def _async_run_job(
    idx: int,
    job: SimulationJob,
    energyplus: EnergyPlusConfig | None,
    cache: SimulationCache | None,
    fs: FileSystem | AsyncFileSystem | None,
    on_progress: Callable[[SimulationProgress], Any] | None,
) -> SimulationResult:
    """Execute a single simulation job, catching SimulationError."""
    # Wrap the user callback to inject batch job context
    job_cb: Callable[[SimulationProgress], Any] | None = None
    if on_progress is not None:
        job_cb = _make_async_job_progress_callback(on_progress, idx, job.label)

    try:
        return await async_simulate(
            model=job.model,  # type: ignore[arg-type]
            weather=job.weather,
            output_dir=job.output_dir,
            energyplus=energyplus,
            expand_objects=job.expand_objects,
            annual=job.annual,
            design_day=job.design_day,
            output_prefix=job.output_prefix,
            output_suffix=job.output_suffix,
            readvars=job.readvars,
            timeout=job.timeout,
            extra_args=list(job.extra_args) if job.extra_args else None,
            cache=cache,
            fs=fs,
            on_progress=job_cb,
        )
    except Exception as exc:
        # Catch all exceptions (SimulationError, ExpandObjectsError,
        # EnergyPlusNotFoundError, etc.) so that a single job failure
        # never crashes the entire batch.
        failed_run_dir = Path(job.output_dir) if job.output_dir is not None else Path(tempfile.mkdtemp())
        exit_code = getattr(exc, "exit_code", None)
        stderr = getattr(exc, "stderr", None) or str(exc)
        return SimulationResult(
            run_dir=failed_run_dir,
            success=False,
            exit_code=exit_code,
            stdout="",
            stderr=stderr,
            runtime_seconds=0.0,
            output_prefix=job.output_prefix,
        )


def _make_async_job_progress_callback(
    user_cb: Callable[[SimulationProgress], Any],
    job_index: int,
    job_label: str,
) -> Callable[[SimulationProgress], Any]:
    """Create a wrapper that stamps each event with batch job context.

    Preserves the sync/async nature of the user callback.
    """
    import inspect

    is_async = inspect.iscoroutinefunction(user_cb)

    if is_async:

        async def _async_wrapper(event: SimulationProgress) -> None:
            stamped = SimulationProgress(
                phase=event.phase,
                message=event.message,
                percent=event.percent,
                environment=event.environment,
                warmup_day=event.warmup_day,
                sim_day=event.sim_day,
                sim_total_days=event.sim_total_days,
                job_index=job_index,
                job_label=job_label,
            )
            await user_cb(stamped)

        return _async_wrapper

    def _wrapper(event: SimulationProgress) -> None:
        stamped = SimulationProgress(
            phase=event.phase,
            message=event.message,
            percent=event.percent,
            environment=event.environment,
            warmup_day=event.warmup_day,
            sim_day=event.sim_day,
            sim_total_days=event.sim_total_days,
            job_index=job_index,
            job_label=job_label,
        )
        user_cb(stamped)

    return _wrapper



from __future__ import annotations

from idfkit.simulation import HTMLResult, SimulationResult
from typing import Any

data: Any = ...  # type: ignore[assignment]
html: HTMLResult | None = ...  # type: ignore[assignment]
result: SimulationResult = ...  # type: ignore[assignment]
table: Any = ...  # type: ignore[assignment]
# --8<-- [start:example]
html = result.html
if html is not None:
    # Iterate all tables
    for table in html:
        print(f"{table.title}: {len(table.rows)} rows")

    # eppy-compatible (title, rows) pairs
    for title, rows in html.titletable():
        print(title)

    # Look up a table by title (case-insensitive substring match)
    table = html.tablebyname("Site and Source Energy")
    if table:
        data = table.to_dict()  # {row_key: {col_header: value}}
        print(data)

    # Get all tables from a specific report
    annual = html.tablesbyreport("Annual Building Utility Performance Summary")

    # Access by index
    first = html.tablebyindex(0)
# --8<-- [end:example]



from __future__ import annotations

from idfkit import IDFDocument, IDFObject
from idfkit.schedules import DayType, Interpolation
from idfkit.simulation import FileSystem
from typing import Any

pd: Any = ...  # type: ignore[assignment]


# --8<-- [start:example]
def to_series(
    schedule: IDFObject,
    year: int = 2024,
    freq: str = "h",  # hourly
    start_date: tuple[int, int] = (1, 1),
    end_date: tuple[int, int] = (12, 31),
    document: IDFDocument | None = None,
    day_type: DayType = DayType.NORMAL,
    interpolation: Interpolation = Interpolation.NO,
    fs: FileSystem | None = None,
) -> pd.Series:
    """
    Convert schedule to pandas Series with DatetimeIndex.

    Requires: pandas (optional dependency)
    """


# --8<-- [end:example]



from __future__ import annotations

from idfkit.weather import DesignDayManager

ddm: DesignDayManager = ...  # type: ignore[assignment]
# --8<-- [start:example]
ddm = DesignDayManager("chicago.ddy")
print(ddm.summary())
# --8<-- [end:example]



from __future__ import annotations

lat: float = ...  # type: ignore[assignment]
lon: float = ...  # type: ignore[assignment]
# --8<-- [start:example]
from idfkit.weather import geocode

# Get coordinates for an address
lat, lon = geocode("350 Fifth Avenue, New York, NY")
print(f"Empire State Building: {lat:.4f}, {lon:.4f}")
# --8<-- [end:example]



from __future__ import annotations

from datetime import datetime
from idfkit import IDFObject
from typing import Any

dt: Any = ...  # type: ignore[assignment]
obj: IDFObject = ...  # type: ignore[assignment]


# --8<-- [start:example]
def evaluate_day_hourly(obj: IDFObject, dt: datetime) -> float:
    hour = dt.hour  # 0-23
    field_name = f"Hour {hour + 1}"  # "Hour 1" through "Hour 24"
    return float(obj[field_name])


# --8<-- [end:example]



from __future__ import annotations

from idfkit import IDFDocument
from idfkit.simulation import BatchResult, SimulationJob, SimulationResult, TimeSeriesResult

batch: BatchResult = ...  # type: ignore[assignment]
job: SimulationJob = ...  # type: ignore[assignment]
jobs: list[SimulationJob] = ...  # type: ignore[assignment]
model: IDFDocument = ...  # type: ignore[assignment]
result: SimulationResult = ...  # type: ignore[assignment]
ts: TimeSeriesResult = ...  # type: ignore[assignment]
variant: IDFDocument = ...  # type: ignore[assignment]
# --8<-- [start:example]
import asyncio
from idfkit.simulation import async_simulate_batch, SimulationJob


async def main():
    # Create variants
    jobs = []
    for insulation in [0.05, 0.10, 0.15, 0.20]:
        variant = model.copy()
        variant["Material"]["Insulation"].thickness = insulation
        jobs.append(
            SimulationJob(
                model=variant,
                weather="weather.epw",
                label=f"insulation-{insulation}m",
                design_day=True,
            )
        )

    # Run all variants
    batch = await async_simulate_batch(jobs, max_concurrent=4)

    # Analyze results
    for job, result in zip(jobs, batch):
        if result.success:
            ts = result.sql.get_timeseries(
                "Zone Mean Air Temperature",
                "ZONE 1",
            )
            print(f"{job.label}: Max temp {max(ts.values):.1f}°C")


asyncio.run(main())
# --8<-- [end:example]



"""Tests for SimulationResult lazy properties."""

from __future__ import annotations

import shutil
import sqlite3
from pathlib import Path

import pytest
from conftest import InMemoryFileSystem

from idfkit.simulation.outputs import OutputVariableIndex
from idfkit.simulation.parsers.csv import CSVResult
from idfkit.simulation.parsers.sql import SQLResult
from idfkit.simulation.result import SimulationResult

FIXTURES = Path(__file__).parent / "fixtures" / "simulation"


def _create_minimal_sql(path: Path) -> None:
    """Create a minimal EnergyPlus SQLite database."""
    conn = sqlite3.connect(str(path))
    cur = conn.cursor()
    cur.execute(
        "CREATE TABLE ReportDataDictionary ("
        "  ReportDataDictionaryIndex INTEGER PRIMARY KEY,"
        "  IsMeter INTEGER, Type TEXT, IndexGroup TEXT,"
        "  TimestepType TEXT, KeyValue TEXT, Name TEXT,"
        "  ReportingFrequency TEXT, ScheduleName TEXT, Units TEXT)"
    )
    cur.execute(
        "CREATE TABLE Time ("
        "  TimeIndex INTEGER PRIMARY KEY, Year INTEGER, Month INTEGER,"
        "  Day INTEGER, Hour INTEGER, Minute INTEGER, Dst INTEGER,"
        "  Interval INTEGER, IntervalType INTEGER, SimulationDays INTEGER,"
        "  DayType TEXT, EnvironmentPeriodIndex INTEGER, WarmupFlag INTEGER)"
    )
    cur.execute(
        "CREATE TABLE ReportData ("
        "  ReportDataIndex INTEGER PRIMARY KEY,"
        "  TimeIndex INTEGER, ReportDataDictionaryIndex INTEGER, Value REAL)"
    )
    cur.execute(
        "CREATE TABLE TabularDataWithStrings ("
        "  TabularDataIndex INTEGER PRIMARY KEY, ReportName TEXT,"
        "  ReportForString TEXT, TableName TEXT, RowName TEXT,"
        "  ColumnName TEXT, Units TEXT, Value TEXT)"
    )
    cur.execute(
        "CREATE TABLE EnvironmentPeriods ("
        "  EnvironmentPeriodIndex INTEGER PRIMARY KEY,"
        "  SimulationIndex INTEGER, EnvironmentName TEXT, EnvironmentType INTEGER)"
    )
    cur.execute(
        "INSERT INTO ReportDataDictionary VALUES "
        "(1, 0, 'Zone', 'Facility', 'Zone', '*', "
        "'Site Outdoor Air Drybulb Temperature', 'Hourly', '', 'C')"
    )
    cur.execute("INSERT INTO EnvironmentPeriods VALUES (1, 1, 'RUN PERIOD 1', 3)")
    cur.execute("INSERT INTO Time VALUES (1, 2017, 1, 1, 1, 0, 0, 60, 1, 1, 'Monday', 1, 0)")
    cur.execute("INSERT INTO ReportData VALUES (1, 1, 1, -5.0)")
    conn.commit()
    conn.close()


@pytest.fixture()
def result_dir(tmp_path: Path) -> Path:
    """Create a simulation output directory with all file types."""
    shutil.copy(FIXTURES / "sample.rdd", tmp_path / "eplusout.rdd")
    shutil.copy(FIXTURES / "sample.mdd", tmp_path / "eplusout.mdd")
    shutil.copy(FIXTURES / "sample.csv", tmp_path / "eplusout.csv")
    shutil.copy(FIXTURES / "sample.err", tmp_path / "eplusout.err")
    _create_minimal_sql(tmp_path / "eplusout.sql")
    return tmp_path


@pytest.fixture()
def result(result_dir: Path) -> SimulationResult:
    """Create a SimulationResult from the test directory."""
    return SimulationResult(
        run_dir=result_dir,
        success=True,
        exit_code=0,
        stdout="",
        stderr="",
        runtime_seconds=1.0,
    )


class TestSqlProperty:
    """Tests for SimulationResult.sql."""

    def test_returns_sql_result(self, result: SimulationResult) -> None:
        sql = result.sql
        assert isinstance(sql, SQLResult)

    def test_is_cached(self, result: SimulationResult) -> None:
        sql1 = result.sql
        sql2 = result.sql
        assert sql1 is sql2

    def test_none_when_no_file(self, tmp_path: Path) -> None:
        r = SimulationResult(
            run_dir=tmp_path,
            success=True,
            exit_code=0,
            stdout="",
            stderr="",
            runtime_seconds=0.0,
        )
        assert r.sql is None

    def test_can_query(self, result: SimulationResult) -> None:
        sql = result.sql
        assert sql is not None
        ts = sql.get_timeseries("Site Outdoor Air Drybulb Temperature")
        assert len(ts.values) == 1
        assert ts.values[0] == -5.0


class TestVariablesProperty:
    """Tests for SimulationResult.variables."""

    def test_returns_index(self, result: SimulationResult) -> None:
        variables = result.variables
        assert isinstance(variables, OutputVariableIndex)

    def test_is_cached(self, result: SimulationResult) -> None:
        v1 = result.variables
        v2 = result.variables
        assert v1 is v2

    def test_variables_loaded(self, result: SimulationResult) -> None:
        variables = result.variables
        assert variables is not None
        assert len(variables.variables) == 7
        assert len(variables.meters) == 5

    def test_none_when_no_file(self, tmp_path: Path) -> None:
        r = SimulationResult(
            run_dir=tmp_path,
            success=True,
            exit_code=0,
            stdout="",
            stderr="",
            runtime_seconds=0.0,
        )
        assert r.variables is None


class TestCsvProperty:
    """Tests for SimulationResult.csv."""

    def test_returns_csv_result(self, result: SimulationResult) -> None:
        csv = result.csv
        assert isinstance(csv, CSVResult)

    def test_is_cached(self, result: SimulationResult) -> None:
        c1 = result.csv
        c2 = result.csv
        assert c1 is c2

    def test_columns_loaded(self, result: SimulationResult) -> None:
        csv = result.csv
        assert csv is not None
        assert len(csv.columns) == 2

    def test_none_when_no_file(self, tmp_path: Path) -> None:
        r = SimulationResult(
            run_dir=tmp_path,
            success=True,
            exit_code=0,
            stdout="",
            stderr="",
            runtime_seconds=0.0,
        )
        assert r.csv is None


class TestFromDirectory:
    """Tests for SimulationResult.from_directory()."""

    def test_lazy_properties_work(self, result_dir: Path) -> None:
        result = SimulationResult.from_directory(result_dir)
        assert result.sql is not None
        assert result.variables is not None
        assert result.csv is not None


# ---------------------------------------------------------------------------
# FileSystem integration tests
# ---------------------------------------------------------------------------


def _populate_memory_fs(fs: InMemoryFileSystem, run_dir: str, tmp_path: Path) -> None:
    """Load fixture files into an InMemoryFileSystem."""
    fixtures = Path(__file__).parent / "fixtures" / "simulation"
    for name, fixture_name in [
        ("eplusout.rdd", "sample.rdd"),
        ("eplusout.mdd", "sample.mdd"),
        ("eplusout.csv", "sample.csv"),
        ("eplusout.err", "sample.err"),
    ]:
        data = (fixtures / fixture_name).read_bytes()
        fs.write_bytes(f"{run_dir}/{name}", data)

    # SQL needs a real local file — write it to the fs as raw bytes
    sql_path = tmp_path / "temp.sql"
    _create_minimal_sql(sql_path)
    fs.write_bytes(f"{run_dir}/{name}", sql_path.read_bytes())
    fs.write_bytes(f"{run_dir}/eplusout.sql", sql_path.read_bytes())


class TestFsIntegration:
    """Tests for SimulationResult with an InMemoryFileSystem backend."""

    @pytest.fixture
    def mem_fs(self, tmp_path: Path) -> tuple[InMemoryFileSystem, str]:
        fs = InMemoryFileSystem()
        run_dir = "output/sim1"
        _populate_memory_fs(fs, run_dir, tmp_path)
        return fs, run_dir

    def test_from_directory_with_fs(self, mem_fs: tuple[InMemoryFileSystem, str]) -> None:
        fs, run_dir = mem_fs
        result = SimulationResult.from_directory(run_dir, fs=fs)
        assert result.fs is fs
        assert result.run_dir == Path(run_dir)

    def test_errors_via_fs(self, mem_fs: tuple[InMemoryFileSystem, str]) -> None:
        fs, run_dir = mem_fs
        result = SimulationResult.from_directory(run_dir, fs=fs)
        report = result.errors
        assert report is not None
        # The sample.err fixture should parse without errors
        assert report.raw_text != ""

    def test_csv_via_fs(self, mem_fs: tuple[InMemoryFileSystem, str]) -> None:
        fs, run_dir = mem_fs
        result = SimulationResult.from_directory(run_dir, fs=fs)
        csv = result.csv
        assert csv is not None
        assert isinstance(csv, CSVResult)
        assert len(csv.columns) == 2

    def test_variables_via_fs(self, mem_fs: tuple[InMemoryFileSystem, str]) -> None:
        fs, run_dir = mem_fs
        result = SimulationResult.from_directory(run_dir, fs=fs)
        variables = result.variables
        assert variables is not None
        assert isinstance(variables, OutputVariableIndex)
        assert len(variables.variables) == 7
        assert len(variables.meters) == 5

    def test_sql_via_fs_downloads_to_temp(self, mem_fs: tuple[InMemoryFileSystem, str]) -> None:
        fs, run_dir = mem_fs
        result = SimulationResult.from_directory(run_dir, fs=fs)
        sql = result.sql
        assert sql is not None
        assert isinstance(sql, SQLResult)
        ts = sql.get_timeseries("Site Outdoor Air Drybulb Temperature")
        assert len(ts.values) == 1
        assert ts.values[0] == -5.0

    def test_find_output_file_via_fs(self, mem_fs: tuple[InMemoryFileSystem, str]) -> None:
        fs, run_dir = mem_fs
        result = SimulationResult.from_directory(run_dir, fs=fs)
        assert result.err_path is not None
        assert result.csv_path is not None
        assert result.rdd_path is not None
        assert result.mdd_path is not None
        assert result.sql_path is not None

    def test_find_output_file_via_fs_none_when_missing(self) -> None:
        fs = InMemoryFileSystem()
        result = SimulationResult.from_directory("empty/dir", fs=fs)
        assert result.err_path is None
        assert result.csv_path is None



from __future__ import annotations

from idfkit import IDFDocument
from idfkit.simulation import SimulationCache

cache: SimulationCache = ...  # type: ignore[assignment]
model: IDFDocument = ...  # type: ignore[assignment]
weather: str = ...  # type: ignore[assignment]
# --8<-- [start:example]
from idfkit.simulation import simulate, SimulationCache

cache = SimulationCache()

# First run: executes EnergyPlus
result1 = simulate(model, weather, cache=cache)

# Second run: instant cache hit
result2 = simulate(model, weather, cache=cache)
# --8<-- [end:example]



"""Download EPW, DDY, and related weather files from climate.onebuilding.org."""

from __future__ import annotations

import logging
import shutil
import time
import zipfile
from dataclasses import dataclass
from datetime import timedelta
from pathlib import Path
from typing import TYPE_CHECKING
from urllib.error import HTTPError, URLError
from urllib.request import Request, urlopen

from .index import default_cache_dir
from .station import WeatherStation

if TYPE_CHECKING:
    from .index import StationIndex

logger = logging.getLogger(__name__)

_USER_AGENT = "idfkit (https://github.com/idfkit/idfkit)"


@dataclass(frozen=True)
class WeatherFiles:
    """Paths to downloaded and extracted weather files.

    Attributes:
        epw: Path to the ``.epw`` file (always present after extraction).
        ddy: Path to the ``.ddy`` file (always present after extraction).
        stat: Path to the ``.stat`` file, or ``None`` if not included.
        zip_path: Path to the original downloaded ZIP archive.
        station: The station this download corresponds to.
    """

    epw: Path
    ddy: Path
    stat: Path | None
    zip_path: Path
    station: WeatherStation


class WeatherDownloader:
    """Download and cache weather files from climate.onebuilding.org.

    Downloaded ZIP archives are extracted and cached locally so that
    subsequent requests for the same station and dataset are served from
    disk without a network call.

    Examples:
        ```python
        from idfkit.weather import StationIndex, WeatherDownloader

        station = StationIndex.load().search("chicago ohare")[0].station
        downloader = WeatherDownloader()
        files = downloader.download(station)
        print(files.epw)
        ```

    Args:
        cache_dir: Override the default cache directory.
        max_age: Maximum age of cached files before re-downloading.
            Can be a [timedelta][datetime.timedelta] or a number of seconds.
            If ``None`` (default), cached files never expire.

    Note:
        The cache has no size limit. For CI/CD environments with limited disk
        space, consider using [clear_cache][idfkit.weather.download.WeatherDownloader.clear_cache] periodically or setting
        a ``max_age`` to force re-downloads of stale files.
    """

    __slots__ = ("_cache_dir", "_max_age_seconds")

    def __init__(
        self,
        cache_dir: Path | None = None,
        max_age: timedelta | float | None = None,
    ) -> None:
        self._cache_dir = cache_dir or default_cache_dir()
        if max_age is None:
            self._max_age_seconds: float | None = None
        elif isinstance(max_age, timedelta):
            self._max_age_seconds = max_age.total_seconds()
        else:
            self._max_age_seconds = float(max_age)

    def _is_stale(self, path: Path) -> bool:
        """Check if a cached file is older than max_age."""
        if self._max_age_seconds is None:
            return False
        if not path.exists():
            return True
        age = time.time() - path.stat().st_mtime
        return age > self._max_age_seconds

    def download(self, station: WeatherStation) -> WeatherFiles:
        """Download and extract weather files for *station*.

        If the files are already cached and not stale, no network request is made.

        Args:
            station: The weather station to download files for.

        Returns:
            A [WeatherFiles][idfkit.weather.download.WeatherFiles] with paths to the extracted files.

        Raises:
            RuntimeError: If the download or extraction fails.
        """
        # Derive a cache subdirectory from the ZIP filename
        zip_filename = station.url.rsplit("/", maxsplit=1)[-1]
        stem = zip_filename.removesuffix(".zip")
        station_dir = self._cache_dir / "files" / str(station.wmo) / stem
        zip_path = station_dir / zip_filename

        # Download if not cached or if stale
        if not zip_path.exists() or self._is_stale(zip_path):
            station_dir.mkdir(parents=True, exist_ok=True)
            logger.info("Downloading weather data for %s (WMO %s)", station.display_name, station.wmo)
            try:
                req = Request(station.url, headers={"User-Agent": _USER_AGENT})  # noqa: S310
                with urlopen(req, timeout=120) as resp:  # noqa: S310
                    zip_path.write_bytes(resp.read())
            except (HTTPError, URLError, TimeoutError, OSError) as exc:
                msg = f"Failed to download weather data from {station.url}: {exc}"
                raise RuntimeError(msg) from exc
        else:
            logger.debug("Cache hit for station %s (WMO %s)", station.display_name, station.wmo)

        # Extract if EPW doesn't already exist or if the ZIP is newer than
        # the EPW (i.e. we just re-downloaded).  We compare against the ZIP's
        # mtime rather than calling ``_is_stale(epw_path)`` because
        # ``zipfile.extractall`` preserves archive-internal timestamps, so the
        # extracted EPW's mtime can be arbitrarily old and would always appear
        # stale.
        epw_path = self._find_file(station_dir, ".epw")
        needs_extract = epw_path is None or (zip_path.exists() and epw_path.stat().st_mtime < zip_path.stat().st_mtime)
        if needs_extract:
            try:
                with zipfile.ZipFile(zip_path) as zf:
                    zf.extractall(station_dir)
            except zipfile.BadZipFile as exc:
                msg = f"Downloaded file is not a valid ZIP archive: {zip_path}"
                raise RuntimeError(msg) from exc
            epw_path = self._find_file(station_dir, ".epw")

        if epw_path is None:
            msg = f"No .epw file found in downloaded archive for {station.display_name}"
            raise RuntimeError(msg)

        ddy_path = self._find_file(station_dir, ".ddy")
        if ddy_path is None:
            msg = f"No .ddy file found in downloaded archive for {station.display_name}"
            raise RuntimeError(msg)

        stat_path = self._find_file(station_dir, ".stat")

        return WeatherFiles(
            epw=epw_path,
            ddy=ddy_path,
            stat=stat_path,
            zip_path=zip_path,
            station=station,
        )

    def get_epw(self, station: WeatherStation) -> Path:
        """Download and return the path to the EPW file."""
        return self.download(station).epw

    def get_ddy(self, station: WeatherStation) -> Path:
        """Download and return the path to the DDY file."""
        return self.download(station).ddy

    def _resolve_filename(self, filename: str, index: StationIndex | None) -> WeatherStation:
        """Resolve an EPW filename to a station, raising on failure."""
        if index is None:
            from .index import StationIndex as _StationIndex

            index = _StationIndex.load()
        stations = index.get_by_filename(filename)
        if not stations:
            msg = f"No weather station found for filename: {filename!r}"
            raise ValueError(msg)
        return stations[0]

    def get_epw_by_filename(
        self,
        filename: str,
        *,
        index: StationIndex | None = None,
    ) -> Path:
        """Download and return the EPW path for an EPW filename.

        Resolves the canonical EPW filename to a station via
        [StationIndex.get_by_filename][idfkit.weather.index.StationIndex.get_by_filename]
        and downloads the corresponding weather files.

        Args:
            filename: EPW filename or stem (with or without extension).
            index: A pre-loaded station index.  If ``None``, loads the
                default index via
                [StationIndex.load][idfkit.weather.index.StationIndex.load].

        Raises:
            ValueError: If the filename does not match any station.
        """
        return self.get_epw(self._resolve_filename(filename, index))

    def get_ddy_by_filename(
        self,
        filename: str,
        *,
        index: StationIndex | None = None,
    ) -> Path:
        """Download and return the DDY path for an EPW filename.

        Same as
        [get_epw_by_filename][idfkit.weather.download.WeatherDownloader.get_epw_by_filename]
        but returns the DDY file path.

        Args:
            filename: EPW filename or stem (with or without extension).
            index: A pre-loaded station index.

        Raises:
            ValueError: If the filename does not match any station.
        """
        return self.get_ddy(self._resolve_filename(filename, index))

    def clear_cache(self) -> None:
        """Remove all cached weather files.

        This removes the entire ``files/`` subdirectory within the cache,
        which contains all downloaded ZIP archives and extracted files.
        """
        files_dir = self._cache_dir / "files"
        if files_dir.exists():
            shutil.rmtree(files_dir)

    @staticmethod
    def _find_file(directory: Path, suffix: str) -> Path | None:
        """Find the first file with the given suffix in *directory*."""
        for p in directory.iterdir():
            if p.suffix.lower() == suffix.lower() and p.is_file():
                return p
        return None



from __future__ import annotations

from idfkit import IDFDocument

model: IDFDocument = ...  # type: ignore[assignment]
# --8<-- [start:example]
# Literal key — fully typed, IDE knows the return type
zones = model["Zone"]

# Dynamic key — use get_collection() for variable object types
obj_type = "Zone"
collection = model.get_collection(obj_type)  # IDFCollection[IDFObject]
# --8<-- [end:example]



from __future__ import annotations

from asyncio import asyncio
from idfkit import IDFDocument
from idfkit.simulation import async_simulate

model: IDFDocument = ...  # type: ignore[assignment]


# --8<-- [start:example]
async def run_with_timeout():
    task = asyncio.create_task(async_simulate(model, "weather.epw"))

    try:
        result = await asyncio.wait_for(task, timeout=120)
    except asyncio.TimeoutError:
        print("Simulation cancelled after 120s")


# --8<-- [end:example]



from __future__ import annotations

from idfkit import IDFDocument

model: IDFDocument = ...  # type: ignore[assignment]
# --8<-- [start:example]
import ipywidgets as widgets
from IPython.display import display
from idfkit.simulation import simulate, SimulationProgress

bar = widgets.FloatProgress(min=0, max=100, description="Simulating:")
label = widgets.Label(value="Starting...")
display(widgets.HBox([bar, label]))


def on_progress(event: SimulationProgress) -> None:
    if event.percent is not None:
        bar.value = event.percent
    label.value = f"{event.phase}: {event.message[:60]}"


result = simulate(model, "weather.epw", annual=True, on_progress=on_progress)
bar.value = 100
label.value = "Done!"
# --8<-- [end:example]



from __future__ import annotations

from idfkit.simulation import SimulationJob, simulate_batch

jobs: list[SimulationJob] = ...  # type: ignore[assignment]
# --8<-- [start:example]
# Use all CPUs
batch = simulate_batch(jobs, max_workers=None)  # Default

# Limit to 4 concurrent simulations
batch = simulate_batch(jobs, max_workers=4)

# Sequential (useful for debugging)
batch = simulate_batch(jobs, max_workers=1)
# --8<-- [end:example]



from __future__ import annotations

from idfkit import IDFDocument
from idfkit.simulation import SimulationResult

model: IDFDocument = ...  # type: ignore[assignment]
result: SimulationResult = ...  # type: ignore[assignment]
# --8<-- [start:example]
import asyncio
from idfkit import load_idf
from idfkit.simulation import async_simulate


async def main():
    model = load_idf("building.idf")
    result = await async_simulate(model, "weather.epw", design_day=True)

    print(f"Success: {result.success}")
    print(f"Runtime: {result.runtime_seconds:.1f}s")


asyncio.run(main())
# --8<-- [end:example]



#!/usr/bin/env python3
"""Build the bundled weather station index.

Downloads all 10 regional Excel index files from climate.onebuilding.org,
parses them, and writes a compressed JSON index to
``src/idfkit/weather/data/stations.json.gz``.

Requires ``openpyxl`` (install via ``pip install idfkit[weather]``).
"""

from __future__ import annotations

import sys
from pathlib import Path

# Ensure the source tree is importable when running from the repo root.
sys.path.insert(0, str(Path(__file__).resolve().parent.parent / "src"))

from idfkit.weather.index import (
    _INDEX_FILES,
    _SOURCES_BASE_URL,
    _download_file,
    _parse_excel,
    _save_compressed_index,
)
from idfkit.weather.station import WeatherStation

_DEST = Path(__file__).resolve().parent.parent / "src" / "idfkit" / "weather" / "data" / "stations.json.gz"


def main() -> None:
    """Download Excel indexes, parse stations, and write compressed index."""
    import tempfile

    with tempfile.TemporaryDirectory() as tmpdir:
        tmp = Path(tmpdir)
        all_stations: list[WeatherStation] = []
        last_modified: dict[str, str] = {}

        for fname in _INDEX_FILES:
            url = f"{_SOURCES_BASE_URL}/{fname}"
            dest = tmp / fname
            print(f"Downloading {fname}...")
            lm = _download_file(url, dest)
            if lm is not None:
                last_modified[fname] = lm

            print(f"  Parsing {fname}...")
            stations = _parse_excel(dest)
            print(f"  Found {len(stations)} stations")
            all_stations.extend(stations)

        print(f"\nTotal stations: {len(all_stations)}")
        print(f"Writing {_DEST}...")
        _save_compressed_index(all_stations, last_modified, _DEST)
        print("Done.")


if __name__ == "__main__":
    main()



"""Schedule:Compact parser and evaluator.

Handles the Schedule:Compact DSL which defines schedules in a compact format
using Through:, For:, Until:, and Interpolate: keywords.
"""

from __future__ import annotations

import re
import weakref
from dataclasses import dataclass
from datetime import date, datetime, time
from typing import TYPE_CHECKING

from idfkit.schedules.day_types import get_applicable_day_types
from idfkit.schedules.time_utils import END_OF_DAY, evaluate_time_values
from idfkit.schedules.types import (
    DAY_TYPE_ALL_OTHER_DAYS,
    DAY_TYPE_ALLDAYS,
    DAY_TYPE_CUSTOM_DAY_1,
    DAY_TYPE_CUSTOM_DAY_2,
    DAY_TYPE_FRIDAY,
    DAY_TYPE_HOLIDAY,
    DAY_TYPE_MONDAY,
    DAY_TYPE_SATURDAY,
    DAY_TYPE_SUMMER_DESIGN,
    DAY_TYPE_SUNDAY,
    DAY_TYPE_THURSDAY,
    DAY_TYPE_TUESDAY,
    DAY_TYPE_WEDNESDAY,
    DAY_TYPE_WEEKDAYS,
    DAY_TYPE_WEEKENDS,
    DAY_TYPE_WINTER_DESIGN,
    CompactDayRule,
    CompactPeriod,
    DayType,
    Interpolation,
    TimeValue,
)

if TYPE_CHECKING:
    from idfkit.objects import IDFObject

# Regex patterns for parsing Schedule:Compact fields
_THROUGH_PATTERN = re.compile(r"^through:\s*(\d{1,2})[/\-](\d{1,2})$", re.IGNORECASE)
_FOR_PATTERN = re.compile(r"^for:\s*(.+)$", re.IGNORECASE)
_UNTIL_PATTERN = re.compile(r"^until:\s*(\d{1,2}):(\d{2})(?::(\d{2}))?$", re.IGNORECASE)
_INTERPOLATE_PATTERN = re.compile(r"^interpolate:\s*(yes|no|average|linear)$", re.IGNORECASE)
_VALUE_PATTERN = re.compile(r"^-?\d+\.?\d*$")

# Day type name mapping (case-insensitive)
_DAY_TYPE_MAP = {
    "sunday": DAY_TYPE_SUNDAY,
    "monday": DAY_TYPE_MONDAY,
    "tuesday": DAY_TYPE_TUESDAY,
    "wednesday": DAY_TYPE_WEDNESDAY,
    "thursday": DAY_TYPE_THURSDAY,
    "friday": DAY_TYPE_FRIDAY,
    "saturday": DAY_TYPE_SATURDAY,
    "weekdays": DAY_TYPE_WEEKDAYS,
    "weekends": DAY_TYPE_WEEKENDS,
    "alldays": DAY_TYPE_ALLDAYS,
    "holidays": DAY_TYPE_HOLIDAY,
    "holiday": DAY_TYPE_HOLIDAY,
    "summerdesignday": DAY_TYPE_SUMMER_DESIGN,
    "winterdesignday": DAY_TYPE_WINTER_DESIGN,
    "customday1": DAY_TYPE_CUSTOM_DAY_1,
    "customday2": DAY_TYPE_CUSTOM_DAY_2,
    "allotherdays": DAY_TYPE_ALL_OTHER_DAYS,
}

# Cache for parse_compact results, keyed by IDFObject identity.
# IDFObject.__hash__ uses id(self) so WeakKeyDictionary works correctly.
# Each entry stores (result_tuple, mutation_version) so that stale
# results are discarded when the object is mutated.
_ParseResult = tuple[list[CompactPeriod], Interpolation]
_parse_cache: weakref.WeakKeyDictionary[IDFObject, tuple[_ParseResult, int]] = weakref.WeakKeyDictionary()


@dataclass
class _ParseState:
    """Internal state for parsing Schedule:Compact."""

    periods: list[CompactPeriod]
    interpolation: Interpolation
    current_period: CompactPeriod | None
    current_rule: CompactDayRule | None
    field_index: int
    consecutive_none: int


def _process_through(state: _ParseState, match: re.Match[str]) -> None:
    """Process a Through: keyword."""
    if state.current_period is not None:
        if state.current_rule is not None:
            state.current_period.day_rules.append(state.current_rule)
        state.periods.append(state.current_period)

    month = int(match.group(1))
    day = int(match.group(2))
    state.current_period = CompactPeriod(end_month=month, end_day=day, day_rules=[])
    state.current_rule = None


def _process_for(state: _ParseState, match: re.Match[str]) -> None:
    """Process a For: keyword."""
    if state.current_rule is not None and state.current_period is not None:
        state.current_period.day_rules.append(state.current_rule)

    day_types_str = match.group(1)
    day_types = _parse_day_types(day_types_str)
    state.current_rule = CompactDayRule(day_types=day_types, time_values=[])


def _process_until(state: _ParseState, match: re.Match[str], ext_fields: list[str]) -> None:
    """Process an Until: keyword."""
    hour = int(match.group(1))
    minute = int(match.group(2))
    second = int(match.group(3)) if match.group(3) else 0

    until_time = END_OF_DAY if hour == 24 else time(hour, minute, second)

    # Next field should be the value
    state.field_index += 1
    if state.field_index < len(ext_fields) and state.current_rule is not None:
        value_str = ext_fields[state.field_index]
        if value_str:
            schedule_value = float(value_str)
            state.current_rule.time_values.append(TimeValue(until_time=until_time, value=schedule_value))


def _process_field(state: _ParseState, value_str: str, ext_fields: list[str]) -> None:
    """Process a single field value in Schedule:Compact."""
    if match := _THROUGH_PATTERN.match(value_str):
        _process_through(state, match)
    elif match := _FOR_PATTERN.match(value_str):
        _process_for(state, match)
    elif match := _UNTIL_PATTERN.match(value_str):
        _process_until(state, match, ext_fields)
    elif (match := _INTERPOLATE_PATTERN.match(value_str)) and match.group(1).lower() in ("yes", "average", "linear"):
        state.interpolation = Interpolation.AVERAGE


def _finalize_parse_state(state: _ParseState) -> None:
    """Save the final period and rule from parse state."""
    if state.current_period is not None:
        if state.current_rule is not None:
            state.current_period.day_rules.append(state.current_rule)
        state.periods.append(state.current_period)


def parse_compact(obj: IDFObject) -> tuple[list[CompactPeriod], Interpolation]:
    """Parse a Schedule:Compact object into structured data.

    Uses ``field_order`` to read extensible fields in their stored sequence,
    avoiding any dependency on naming conventions (``field`` vs ``field_1``).

    Results are cached per object identity (the cache uses a WeakKeyDictionary
    so entries are automatically cleaned up when the object is garbage-collected).

    Args:
        obj: The Schedule:Compact object.

    Returns:
        Tuple of (list of CompactPeriod, interpolation mode).

    Raises:
        ValueError: If the schedule syntax is invalid.
    """
    cached = _parse_cache.get(obj)
    if cached is not None:
        result, version = cached
        if version == obj.mutation_version:
            return result

    # Build the ordered list of extensible field values from field_order.
    # The first entry is always schedule_type_limits_name (non-extensible);
    # everything after it is compact DSL data.
    ext_fields: list[str] = []
    field_order = obj.field_order or []
    for name in field_order[1:]:
        raw = obj.data.get(name)
        ext_fields.append(str(raw).strip() if raw is not None else "")

    state = _ParseState(
        periods=[],
        interpolation=Interpolation.NO,
        current_period=None,
        current_rule=None,
        field_index=0,
        consecutive_none=0,
    )

    while state.field_index < len(ext_fields):
        value_str = ext_fields[state.field_index]

        if not value_str:
            state.consecutive_none += 1
            if state.consecutive_none >= 3:
                break
            state.field_index += 1
            continue

        state.consecutive_none = 0
        _process_field(state, value_str, ext_fields)
        state.field_index += 1

    _finalize_parse_state(state)
    result: _ParseResult = (state.periods, state.interpolation)
    _parse_cache[obj] = (result, obj.mutation_version)
    return result


def _parse_day_types(day_types_str: str) -> set[str]:
    """Parse a day types string from the For: field.

    Can contain multiple day types separated by spaces or commas.

    Args:
        day_types_str: The day types string.

    Returns:
        Set of normalized day type strings.
    """
    day_types: set[str] = set()

    # Split by whitespace and commas
    parts = re.split(r"[\s,]+", day_types_str)

    for part in parts:
        part = part.strip().lower()
        if part in _DAY_TYPE_MAP:
            day_types.add(_DAY_TYPE_MAP[part])

    return day_types


def evaluate_compact(
    obj: IDFObject,
    dt: datetime,
    day_type: DayType = DayType.NORMAL,
    holidays: set[date] | None = None,
    custom_day_1: set[date] | None = None,
    custom_day_2: set[date] | None = None,
) -> float:
    """Evaluate a Schedule:Compact at a specific datetime.

    Args:
        obj: The Schedule:Compact object.
        dt: The datetime to evaluate.
        day_type: Override day type.
        holidays: Set of holiday dates.
        custom_day_1: Set of CustomDay1 dates.
        custom_day_2: Set of CustomDay2 dates.

    Returns:
        The schedule value.
    """
    periods, interpolation = parse_compact(obj)

    if not periods:
        return 0.0

    d = dt.date()
    current_time = dt.time()

    # Find the period containing this date
    period = _find_period_for_date(periods, d)
    if period is None:
        return 0.0

    # Get applicable day types for this date
    applicable_types = get_applicable_day_types(
        d, day_type, holidays or set(), custom_day_1 or set(), custom_day_2 or set()
    )

    # Find the matching day rule
    rule = _find_matching_rule(period.day_rules, applicable_types)
    if rule is None:
        return 0.0

    # Evaluate the time-value pairs
    return evaluate_time_values(rule.time_values, current_time, interpolation)


def _find_period_for_date(periods: list[CompactPeriod], d: date) -> CompactPeriod | None:
    """Find the period containing a date.

    Periods are sequential and cover the entire year. Each period's end date
    is its boundary, and the next period starts the following day.

    Args:
        periods: List of periods in order.
        d: The date to find.

    Returns:
        The period containing the date, or None.
    """
    for period in periods:
        if (d.month, d.day) <= (period.end_month, period.end_day):
            return period

    # If date is past all periods, use the last period (wrap around to next year)
    return periods[-1] if periods else None


def _find_matching_rule(rules: list[CompactDayRule], applicable_types: set[str]) -> CompactDayRule | None:
    """Find the first rule that matches the applicable day types.

    Rules are checked in order. The first rule with a matching day type wins.
    More specific day types (like individual weekdays) should be listed before
    more general ones (like AllDays).

    Args:
        rules: List of day rules.
        applicable_types: Set of applicable day type strings.

    Returns:
        The first matching rule, or None.
    """
    from idfkit.schedules.day_types import DAY_TYPE_PRIORITY

    # For each priority level, check if any rule matches
    for priority_type in DAY_TYPE_PRIORITY:
        if priority_type not in applicable_types:
            continue

        for rule in rules:
            if priority_type in rule.day_types:
                return rule

    # Fallback: check AllOtherDays
    for rule in rules:
        if DAY_TYPE_ALL_OTHER_DAYS in rule.day_types:
            return rule

    return None



from __future__ import annotations

from idfkit.weather import WeatherDownloader, WeatherStation

downloader: WeatherDownloader = ...  # type: ignore[assignment]
station: WeatherStation = ...  # type: ignore[assignment]
# --8<-- [start:example]
from idfkit.weather import WeatherDownloader

downloader = WeatherDownloader()

try:
    files = downloader.download(station)
except Exception as e:
    print(f"Download failed: {e}")
# --8<-- [end:example]



"""Tests for ReferenceGraph class."""

from __future__ import annotations

from idfkit.objects import IDFObject
from idfkit.references import ReferenceGraph


class TestReferenceGraphRegister:
    def test_register_single(self) -> None:
        graph = ReferenceGraph()
        obj = IDFObject(obj_type="People", name="P1", data={"zone_name": "Z1"})
        graph.register(obj, "zone_name", "Z1")
        refs = graph.get_referencing("Z1")
        assert obj in refs

    def test_register_empty_name_ignored(self) -> None:
        graph = ReferenceGraph()
        obj = IDFObject(obj_type="People", name="P1")
        graph.register(obj, "zone_name", "")
        assert len(graph) == 0

    def test_register_multiple_references(self) -> None:
        graph = ReferenceGraph()
        obj = IDFObject(obj_type="People", name="P1")
        graph.register(obj, "zone_name", "Z1")
        graph.register(obj, "schedule_name", "S1")
        refs = graph.get_references(obj)
        assert "Z1" in refs
        assert "S1" in refs

    def test_register_object_list(self) -> None:
        graph = ReferenceGraph()
        graph.register_object_list("ZoneNames", "Zone")
        stats = graph.stats()
        assert stats["object_lists"] == 1


class TestReferenceGraphLookup:
    def test_get_referencing(self, reference_graph: ReferenceGraph) -> None:
        refs = reference_graph.get_referencing("Zone1")
        assert len(refs) == 2

    def test_get_referencing_case_insensitive(self, reference_graph: ReferenceGraph) -> None:
        refs1 = reference_graph.get_referencing("Zone1")
        refs2 = reference_graph.get_referencing("zone1")
        refs3 = reference_graph.get_referencing("ZONE1")
        assert refs1 == refs2 == refs3

    def test_get_referencing_nonexistent(self, reference_graph: ReferenceGraph) -> None:
        refs = reference_graph.get_referencing("DoesNotExist")
        assert len(refs) == 0

    def test_get_referencing_with_fields(self, reference_graph: ReferenceGraph) -> None:
        refs = reference_graph.get_referencing_with_fields("Zone1")
        assert len(refs) == 2
        for _obj, field in refs:
            assert field == "zone_name"

    def test_get_references(self) -> None:
        graph = ReferenceGraph()
        obj = IDFObject(obj_type="People", name="P1")
        graph.register(obj, "zone_name", "Z1")
        graph.register(obj, "schedule_name", "S1")
        refs = graph.get_references(obj)
        assert refs == {"Z1", "S1"}

    def test_get_references_with_fields(self) -> None:
        graph = ReferenceGraph()
        obj = IDFObject(obj_type="People", name="P1")
        graph.register(obj, "zone_name", "Z1")
        refs = graph.get_references_with_fields(obj)
        assert ("Z1", "zone_name") in refs

    def test_get_references_unknown_object(self) -> None:
        graph = ReferenceGraph()
        obj = IDFObject(obj_type="People", name="P1")
        refs = graph.get_references(obj)
        assert len(refs) == 0

    def test_is_referenced(self, reference_graph: ReferenceGraph) -> None:
        assert reference_graph.is_referenced("Zone1") is True
        assert reference_graph.is_referenced("Zone2") is True
        assert reference_graph.is_referenced("Zone3") is False


class TestReferenceGraphUnregister:
    def test_unregister(self) -> None:
        graph = ReferenceGraph()
        obj = IDFObject(obj_type="People", name="P1")
        graph.register(obj, "zone_name", "Z1")
        graph.unregister(obj)
        assert len(graph.get_referencing("Z1")) == 0
        assert len(graph.get_references(obj)) == 0

    def test_unregister_cleans_empty_refs(self) -> None:
        graph = ReferenceGraph()
        obj = IDFObject(obj_type="People", name="P1")
        graph.register(obj, "zone_name", "Z1")
        graph.unregister(obj)
        assert not graph.is_referenced("Z1")

    def test_unregister_unknown_object(self) -> None:
        graph = ReferenceGraph()
        obj = IDFObject(obj_type="People", name="P1")
        # Should not raise
        graph.unregister(obj)


class TestReferenceGraphDanglingReferences:
    def test_no_dangling(self, reference_graph: ReferenceGraph) -> None:
        valid_names = {"ZONE1", "ZONE2"}
        dangling = list(reference_graph.get_dangling_references(valid_names))
        assert len(dangling) == 0

    def test_with_dangling(self, reference_graph: ReferenceGraph) -> None:
        valid_names = {"ZONE1"}  # Zone2 not valid
        dangling = list(reference_graph.get_dangling_references(valid_names))
        assert len(dangling) == 1
        _obj, field, name = dangling[0]
        assert name == "ZONE2"
        assert field == "zone_name"

    def test_all_dangling(self) -> None:
        graph = ReferenceGraph()
        obj = IDFObject(obj_type="People", name="P1")
        graph.register(obj, "zone_name", "NonexistentZone")
        dangling = list(graph.get_dangling_references(set()))
        assert len(dangling) == 1


class TestReferenceGraphMisc:
    def test_clear(self, reference_graph: ReferenceGraph) -> None:
        reference_graph.clear()
        assert len(reference_graph) == 0
        assert reference_graph.stats()["total_references"] == 0

    def test_len(self, reference_graph: ReferenceGraph) -> None:
        assert len(reference_graph) == 3  # 3 registrations

    def test_stats(self, reference_graph: ReferenceGraph) -> None:
        stats = reference_graph.stats()
        assert stats["total_references"] == 3
        assert stats["objects_with_references"] == 3
        assert stats["names_referenced"] == 2


class TestReferenceGraphRenameTarget:
    def test_rename_target_updates_referenced_by(self) -> None:
        graph = ReferenceGraph()
        obj = IDFObject(obj_type="People", name="P1")
        graph.register(obj, "zone_name", "OldZone")
        graph.rename_target("OldZone", "NewZone")
        assert graph.is_referenced("NewZone")
        assert not graph.is_referenced("OldZone")

    def test_rename_target_updates_references(self) -> None:
        graph = ReferenceGraph()
        obj = IDFObject(obj_type="People", name="P1")
        graph.register(obj, "zone_name", "OldZone")
        graph.rename_target("OldZone", "NewZone")
        refs = graph.get_references(obj)
        assert "NEWZONE" in refs
        assert "OLDZONE" not in refs

    def test_rename_target_multiple_referrers(self) -> None:
        graph = ReferenceGraph()
        obj_a = IDFObject(obj_type="People", name="P1")
        obj_b = IDFObject(obj_type="Lights", name="L1")
        graph.register(obj_a, "zone_name", "Z1")
        graph.register(obj_b, "zone_name", "Z1")
        graph.rename_target("Z1", "Z1_Renamed")
        refs = graph.get_referencing("Z1_Renamed")
        assert obj_a in refs
        assert obj_b in refs
        assert not graph.is_referenced("Z1")

    def test_rename_target_noop_same_name(self) -> None:
        graph = ReferenceGraph()
        obj = IDFObject(obj_type="People", name="P1")
        graph.register(obj, "zone_name", "Z1")
        graph.rename_target("Z1", "Z1")
        assert graph.is_referenced("Z1")

    def test_rename_target_nonexistent_is_noop(self) -> None:
        graph = ReferenceGraph()
        graph.rename_target("NoSuchName", "Whatever")
        assert len(graph) == 0

    def test_rename_target_merges_with_existing(self) -> None:
        graph = ReferenceGraph()
        obj_a = IDFObject(obj_type="People", name="P1")
        obj_b = IDFObject(obj_type="Lights", name="L1")
        graph.register(obj_a, "zone_name", "Z1")
        graph.register(obj_b, "zone_name", "Z2")
        # Rename Z1 -> Z2 should merge
        graph.rename_target("Z1", "Z2")
        refs = graph.get_referencing("Z2")
        assert obj_a in refs
        assert obj_b in refs


class TestReferenceGraphUpdateReference:
    def test_update_reference_basic(self) -> None:
        graph = ReferenceGraph()
        obj = IDFObject(obj_type="People", name="P1")
        graph.register(obj, "zone_name", "Z1")
        graph.update_reference(obj, "zone_name", "Z1", "Z2")
        assert graph.is_referenced("Z2")
        assert not graph.is_referenced("Z1")
        refs = graph.get_references(obj)
        assert "Z2" in refs
        assert "Z1" not in refs

    def test_update_reference_old_none(self) -> None:
        graph = ReferenceGraph()
        obj = IDFObject(obj_type="People", name="P1")
        graph.update_reference(obj, "zone_name", None, "Z1")
        assert graph.is_referenced("Z1")

    def test_update_reference_new_none(self) -> None:
        graph = ReferenceGraph()
        obj = IDFObject(obj_type="People", name="P1")
        graph.register(obj, "zone_name", "Z1")
        graph.update_reference(obj, "zone_name", "Z1", None)
        assert not graph.is_referenced("Z1")

    def test_update_reference_new_empty_string(self) -> None:
        graph = ReferenceGraph()
        obj = IDFObject(obj_type="People", name="P1")
        graph.register(obj, "zone_name", "Z1")
        graph.update_reference(obj, "zone_name", "Z1", "")
        assert not graph.is_referenced("Z1")



from __future__ import annotations

from idfkit import IDFDocument

doc: IDFDocument = ...  # type: ignore[assignment]
# --8<-- [start:example]
from idfkit.simulation import simulate

result = simulate(doc, "weather.epw")
print(result.errors.summary())
# --8<-- [end:example]



"""Eppy-compatibility helpers for [IDFDocument][idfkit.document.IDFDocument].

This module contains methods that exist **solely** to ease migration from
`eppy <https://github.com/santoshphilip/eppy>`_.  For each method there is
a recommended idfkit alternative noted in the docstring.

The mixin is mixed into [IDFDocument][idfkit.document.IDFDocument] so that existing eppy code
continues to work without modification.  In a future release the mixin
may be deprecated and eventually removed.

To opt out early, avoid importing or calling any symbol from this module.
"""

from __future__ import annotations

from collections.abc import Iterator, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any

from .writers import OutputType

if TYPE_CHECKING:
    from .objects import IDFCollection, IDFObject
    from .schema import EpJSONSchema
    from .simulation.result import SimulationResult


# ---------------------------------------------------------------------------
# _IDFObjectsView: case-insensitive dict-like wrapper
# ---------------------------------------------------------------------------


class _IDFObjectsView:
    """Case-insensitive view over document collections (eppy compatibility).

    Allows ``doc.idfobjects["ZONE"]`` to work the same as ``doc["Zone"]``.
    """

    __slots__ = ("_doc",)

    def __init__(self, doc: EppyDocumentMixin) -> None:
        self._doc = doc

    def __getitem__(self, key: str) -> IDFCollection[IDFObject]:
        collections = self._doc.collections
        if key in collections:
            return collections[key]
        key_upper = key.upper()
        for obj_type, collection in collections.items():
            if obj_type.upper() == key_upper:
                return collection
        return self._doc[key]

    def __contains__(self, key: object) -> bool:
        if not isinstance(key, str):
            return False
        collections = self._doc.collections
        key_upper = key.upper()
        for obj_type, collection in collections.items():
            if obj_type.upper() == key_upper:
                return len(collection) > 0
        return False

    def __iter__(self) -> Iterator[str]:
        return iter(self._doc.collections)

    def keys(self) -> list[str]:
        """Return object type names."""
        return list(self._doc.collections.keys())

    def values(self) -> list[IDFCollection[IDFObject]]:
        """Return all collections."""
        return list(self._doc.collections.values())

    def items(self) -> list[tuple[str, IDFCollection[IDFObject]]]:
        """Return (type, collection) pairs."""
        return list(self._doc.collections.items())


# ---------------------------------------------------------------------------
# Mixin
# ---------------------------------------------------------------------------


class EppyDocumentMixin:
    """Eppy-compatibility methods for [IDFDocument][idfkit.document.IDFDocument].

    Mixed into [IDFDocument][idfkit.document.IDFDocument] automatically.  Each method documents
    the preferred idfkit API in a tip admonition.
    """

    __slots__ = ()

    # -- TYPE_CHECKING interface (attributes provided by IDFDocument) ---------
    if TYPE_CHECKING:
        _collections: dict[str, IDFCollection[IDFObject]]
        _schema: EpJSONSchema | None
        filepath: Path | None

        @property
        def collections(self) -> dict[str, IDFCollection[IDFObject]]: ...
        def __getitem__(self, key: str) -> IDFCollection[IDFObject]: ...
        def add(self, obj_type: str, name: str = "", **kwargs: Any) -> IDFObject: ...
        def removeidfobject(self, obj: IDFObject) -> None: ...

        @staticmethod
        def _compute_ref_fields(schema: EpJSONSchema, obj_type: str) -> frozenset[str]: ...
        def _index_object_references(self, obj: IDFObject) -> None: ...

    # -- Object access -------------------------------------------------------

    @property
    def idfobjects(self) -> _IDFObjectsView:
        """Dict-like access to all object collections (eppy compatibility).

        Returns a case-insensitive view so ``idf.idfobjects["ZONE"]``
        continues to work.

        !!! tip
            Prefer bracket access for new code:

                ```python
                zones = doc["Zone"]
                zone = doc["Zone"]["Office"]
                ```
        """
        return _IDFObjectsView(self)

    def getobject(self, obj_type: str, name: str) -> IDFObject | None:
        """Get a specific object by type and name (eppy compatibility).

        !!! tip
            Prefer bracket access for new code:

                ```python
                zone = doc["Zone"]["Office"]
                ```

        Args:
            obj_type: Object type (e.g., "Zone")
            name: Object name

        Returns:
            IDFObject or None if not found

        Examples:
            >>> from idfkit import new_document
            >>> model = new_document()
            >>> model.add("Zone", "Perimeter_ZN_1")  # doctest: +ELLIPSIS
            Zone('Perimeter_ZN_1')
            >>> model.getobject("Zone", "Perimeter_ZN_1").name
            'Perimeter_ZN_1'
            >>> model.getobject("Zone", "NonExistent") is None
            True
        """
        collection = self._collections.get(obj_type)
        if collection:
            return collection.get(name)
        return None

    def getiddgroupdict(self) -> dict[str, list[str]]:
        """Get dict mapping IDD group names to their object types (eppy compatibility).

        Uses actual schema group metadata when a schema is loaded, so the
        returned group names match those in the EnergyPlus IDD (e.g.
        ``"Thermal Zones and Surfaces"``).

        Returns:
            Dict of ``{group_name: [object_type, ...]}`` for every object
            type present in the document.

        Examples:
            >>> from idfkit import new_document
            >>> model = new_document()
            >>> model.add("Zone", "Office")  # doctest: +ELLIPSIS
            Zone('Office')
            >>> groups = model.getiddgroupdict()
            >>> "Office" not in groups  # groups are IDD group names, not object names
            True
        """
        groups: dict[str, list[str]] = {}
        for obj_type in self._collections:
            group: str | None = None
            if self._schema is not None:
                group = self._schema.get_group(obj_type)
            if group is None:
                group = "Miscellaneous"
            if group not in groups:
                groups[group] = []
            groups[group].append(obj_type)
        return groups

    # -- Object creation / mutation ------------------------------------------

    def newidfobject(self, obj_type: str, **kwargs: Any) -> IDFObject:
        """Create a new object (eppy compatibility).

        Provided for migration convenience.  The ``Name`` kwarg becomes
        the object name if provided.

        !!! tip
            Prefer [add][idfkit.document.IDFDocument.add] for new code:

                ```python
                zone = doc.add("Zone", "Office", x_origin=0.0)
                ```
        """
        name = kwargs.pop("Name", kwargs.pop("name", ""))
        return self.add(obj_type, name, **kwargs)

    def addidfobject(self, obj: IDFObject) -> IDFObject:
        """Add an existing IDFObject to the document (eppy compatibility).

        !!! tip
            Prefer [add][idfkit.document.IDFDocument.add] for new code.
        """
        object.__setattr__(obj, "_document", self)

        if object.__getattribute__(obj, "_ref_fields") is None and self._schema:
            object.__setattr__(
                obj,
                "_ref_fields",
                self._compute_ref_fields(self._schema, obj.obj_type),
            )

        self[obj.obj_type].add(obj)
        self._index_object_references(obj)
        return obj

    def addidfobjects(self, objects: list[IDFObject]) -> list[IDFObject]:
        """Add multiple objects to the document (eppy compatibility).

        !!! tip
            Prefer calling [add][idfkit.document.IDFDocument.add] in a loop.
        """
        return [self.addidfobject(obj) for obj in objects]

    def popidfobject(self, obj_type: str, index: int) -> IDFObject:
        """Remove and return an object by type and index (eppy compatibility).

        !!! tip
            Prefer [removeidfobject][idfkit.document.IDFDocument.removeidfobject]
            with a direct object reference.

        Args:
            obj_type: Object type (e.g. "Zone")
            index: Zero-based index within the collection

        Returns:
            The removed IDFObject

        Raises:
            IndexError: If the index is out of range
        """
        collection = self[obj_type]
        obj = collection[index]
        self.removeidfobject(obj)
        return obj

    def removeidfobjects(self, objects: Sequence[IDFObject]) -> None:
        """Remove multiple objects from the document (eppy compatibility).

        !!! tip
            Call [removeidfobject][idfkit.document.IDFDocument.removeidfobject] in a loop.
        """
        for obj in objects:
            self.removeidfobject(obj)

    def removeallidfobjects(self, obj_type: str) -> None:
        """Remove all objects of a given type from the document (eppy compatibility).

        !!! tip
            For new code, iterate and remove directly:

                ```python
                for obj in list(doc[obj_type]):
                    doc.removeidfobject(obj)
                ```

        Args:
            obj_type: Object type to remove all instances of (e.g. ``"Zone"``).

        Examples:
            >>> from idfkit import new_document
            >>> model = new_document()
            >>> model.add("Zone", "A")  # doctest: +ELLIPSIS
            Zone('A')
            >>> model.add("Zone", "B")  # doctest: +ELLIPSIS
            Zone('B')
            >>> model.removeallidfobjects("Zone")
            >>> len(model["Zone"])
            0
        """
        for obj in list(self[obj_type]):
            self.removeidfobject(obj)

    def copyidfobject(self, obj: IDFObject, new_name: str | None = None) -> IDFObject:
        """Create a copy of an object with optional new name (eppy compatibility).

        !!! tip
            Use ``obj.copy()`` to create a detached copy, then
            [addidfobject][] to add it.
        """
        new_obj = obj.copy()
        if new_name:
            new_obj.name = new_name
        return self.addidfobject(new_obj)

    def update(self, updates: dict[str, Any]) -> None:
        """Apply batch field updates using dot-separated key paths (eppy compatibility).

        Mirrors eppy's ``json_functions.updateidf`` for parametric
        sweeps and bulk modifications.

        Each key has the form ``"ObjectType.ObjectName.field_name"``
        and the value is the new field value.  The object must already
        exist in the document.

        !!! tip
            For new code, modify objects directly:

                ```python
                zone = doc["Zone"]["Office"]
                zone.x_origin = 10.0
                ```

        !!! note
            Object names containing literal dots are **not** supported
            because the separator is itself a dot.  For objects whose
            names may contain dots, modify fields directly via
            [getobject][] instead.

        Args:
            updates: Mapping of ``"Type.Name.field"`` -> new_value.

        Raises:
            KeyError: If the referenced object does not exist or the
                key format is invalid.

        Examples:
            Batch-update fields for parametric analysis -- for instance,
            sweep insulation thickness without touching the rest of the model:

            >>> from idfkit import new_document
            >>> model = new_document()
            >>> model.add("Material", "Insulation",
            ...     roughness="Rough", thickness=0.05,
            ...     conductivity=0.04, density=30.0, specific_heat=1500.0)  # doctest: +ELLIPSIS
            Material('Insulation')
            >>> model.update({"Material.Insulation.thickness": 0.1})
            >>> model.getobject("Material", "Insulation").thickness
            0.1
        """
        from .objects import to_python_name

        for dotted_key, value in updates.items():
            parts = dotted_key.split(".", 2)
            if len(parts) != 3:
                msg = f"Expected 'ObjectType.ObjectName.field_name', got '{dotted_key}'"
                raise KeyError(msg)
            obj_type, obj_name, field_name = parts
            obj = self.getobject(obj_type, obj_name)
            if obj is None:
                msg = f"No {obj_type} named '{obj_name}'"
                raise KeyError(msg)
            setattr(obj, to_python_name(field_name), value)

    # -- Surfaces (common access pattern) ------------------------------------

    def getsurfaces(self, surface_type: str | None = None) -> list[IDFObject]:
        """Get building surfaces, optionally filtered by type (eppy compatibility).

        !!! tip
            Prefer filtering directly for new code:

                ```python
                walls = doc["BuildingSurface:Detailed"].filter(
                    lambda s: s.surface_type == "Wall"
                )
                ```

        Args:
            surface_type: Filter by surface type ("wall", "floor", "roof", "ceiling")

        Returns:
            List of surface objects

        Examples:
            Query the building envelope by surface type:

            >>> from idfkit import new_document
            >>> model = new_document()
            >>> model.add("BuildingSurface:Detailed", "South_Wall",
            ...     surface_type="Wall", construction_name="", zone_name="",
            ...     outside_boundary_condition="Outdoors",
            ...     sun_exposure="SunExposed", wind_exposure="WindExposed",
            ...     validate=False)  # doctest: +ELLIPSIS
            BuildingSurface:Detailed('South_Wall')
            >>> len(model.getsurfaces())
            1
            >>> len(model.getsurfaces("wall"))
            1
            >>> len(model.getsurfaces("floor"))
            0
        """
        surfaces: list[IDFObject] = list(self["BuildingSurface:Detailed"])

        if surface_type:
            surface_type_upper = surface_type.upper()
            surfaces = [s for s in surfaces if (getattr(s, "surface_type", None) or "").upper() == surface_type_upper]

        return surfaces

    # -- I/O -----------------------------------------------------------------

    def save(
        self,
        filepath: str | Path | None = None,
        encoding: str = "latin-1",
        output_type: OutputType = "standard",
    ) -> None:
        """Save the document to its current filepath (eppy compatibility).

        !!! tip
            Also the recommended idfkit API.  For format conversion, use
            [write_idf][idfkit.writers.write_idf] or
            [write_epjson][idfkit.writers.write_epjson] directly.

        Args:
            filepath: Explicit path override.  If ``None``, uses
                ``self.filepath``.
            encoding: Output encoding (default ``latin-1``).
            output_type: IDF formatting mode — ``"standard"`` (with field
                comments), ``"nocomment"``, or ``"compressed"`` (single-line).

        Raises:
            ValueError: If no filepath is set and none is provided.

        Examples:
            Save a modified model back to its original IDF file:

                ```python
                model = load_idf("5ZoneAirCooled.idf")
                model.add("Zone", "Mech_Room")
                model.save()   # overwrites 5ZoneAirCooled.idf
                ```

            Save to a new path for archival:

                ```python
                model.save("5ZoneAirCooled_v2.idf")
                ```
        """
        from .writers import write_idf

        target = Path(filepath) if filepath else self.filepath
        if target is None:
            msg = "No filepath set - pass a path or use saveas()"
            raise ValueError(msg)
        write_idf(self, target, encoding=encoding, output_type=output_type)  # type: ignore[arg-type]
        self.filepath = target

    def saveas(
        self,
        filepath: str | Path,
        encoding: str = "latin-1",
        output_type: OutputType = "standard",
    ) -> None:
        """Save to a new path and update ``self.filepath`` (eppy compatibility).

        After saving, ``self.filepath`` is updated to the new path so
        subsequent [save][] calls write to it.

        !!! tip
            Also the recommended idfkit API.

        Args:
            filepath: Destination path.
            encoding: Output encoding.
            output_type: IDF formatting mode — ``"standard"`` (with field
                comments), ``"nocomment"``, or ``"compressed"`` (single-line).

        Examples:
            Save under a new name and continue editing there:

                ```python
                model = load_idf("Baseline.idf")
                model.saveas("HighInsulation_Variant.idf")
                model.save()   # now writes to HighInsulation_Variant.idf
                ```
        """
        from .writers import write_idf

        target = Path(filepath)
        write_idf(self, target, encoding=encoding, output_type=output_type)  # type: ignore[arg-type]
        self.filepath = target

    def savecopy(
        self,
        filepath: str | Path,
        encoding: str = "latin-1",
        output_type: OutputType = "standard",
    ) -> None:
        """Save a copy without changing ``self.filepath`` (eppy compatibility).

        Unlike [saveas][], the document's ``filepath`` remains
        unchanged.

        !!! tip
            Also the recommended idfkit API.

        Args:
            filepath: Destination path.
            encoding: Output encoding.
            output_type: IDF formatting mode — ``"standard"`` (with field
                comments), ``"nocomment"``, or ``"compressed"`` (single-line).

        Examples:
            Create a snapshot before running a parametric sweep:

                ```python
                model = load_idf("Baseline.idf")
                model.savecopy("Baseline_backup.idf")
                # ... modify model ...
                model.save()   # still writes to Baseline.idf
                ```
        """
        from .writers import write_idf

        write_idf(self, Path(filepath), encoding=encoding, output_type=output_type)  # type: ignore[arg-type]

    # -- Simulation ----------------------------------------------------------

    def run(
        self,
        weather: str | Path,
        **kwargs: Any,
    ) -> SimulationResult:
        """Run an EnergyPlus simulation (eppy-compatible convenience method).

        This is a thin wrapper around [idfkit.simulation.simulate][]
        provided for users migrating from eppy's ``idf.run()`` pattern.

        !!! tip
            For new code, prefer the standalone function:

                ```python
                from idfkit.simulation import simulate
                result = simulate(model, "weather.epw")
                ```

            It offers the same parameters and keeps simulation concerns
            separate from document manipulation.

        Args:
            weather: Path to the EPW weather file.
            **kwargs: Forwarded to [simulate][idfkit.simulation.simulate]
                (e.g. ``output_dir``, ``annual``, ``design_day``,
                ``timeout``, ``energyplus``).

        Returns:
            [SimulationResult][idfkit.simulation.SimulationResult] with paths to all
            output files, error reports, and SQL queries.

        Raises:
            EnergyPlusNotFoundError: If no EnergyPlus installation is found.
            SimulationError: If the simulation fails.

        Examples:
            Run a design-day simulation:

                ```python
                model = load_idf("5ZoneAirCooled.idf")
                result = model.run("weather.epw", design_day=True)
                print(result.errors.summary())
                ```
        """
        from .simulation import simulate

        return simulate(self, weather, **kwargs)  # type: ignore[arg-type]



"""Tests for schedules.time_utils module."""

from __future__ import annotations

from datetime import time

from idfkit.schedules.time_utils import END_OF_DAY, END_OF_DAY_MINUTES, evaluate_time_values, time_to_minutes
from idfkit.schedules.types import Interpolation, TimeValue


class TestTimeToMinutes:
    """Tests for time_to_minutes function."""

    def test_midnight(self) -> None:
        """Test midnight returns 0."""
        assert time_to_minutes(time(0, 0)) == 0.0

    def test_noon(self) -> None:
        """Test noon returns 720."""
        assert time_to_minutes(time(12, 0)) == 720.0

    def test_with_minutes(self) -> None:
        """Test time with minutes."""
        assert time_to_minutes(time(8, 30)) == 510.0

    def test_with_seconds(self) -> None:
        """Test time with seconds."""
        assert time_to_minutes(time(8, 0, 30)) == 480.5

    def test_end_of_day_sentinel_returns_exactly_1440(self) -> None:
        """Test END_OF_DAY sentinel returns exactly 1440.0."""
        result = time_to_minutes(END_OF_DAY)
        assert result == END_OF_DAY_MINUTES
        assert result == 1440.0

    def test_end_of_day_is_exact(self) -> None:
        """Test END_OF_DAY produces exact value, not floating-point approximation."""
        # Without sentinel detection, time(23, 59, 59, 999999) would give
        # 1439.999... due to floating-point arithmetic
        result = time_to_minutes(END_OF_DAY)
        assert result == 1440.0
        assert isinstance(result, float)


class TestEvaluateTimeValues:
    """Tests for evaluate_time_values function."""

    def test_empty_returns_zero(self) -> None:
        """Test empty time_values returns 0.0."""
        result = evaluate_time_values([], time(12, 0), Interpolation.NO)
        assert result == 0.0

    def test_step_function(self) -> None:
        """Test step function (no interpolation)."""
        time_values = [
            TimeValue(until_time=time(8, 0), value=0.0),
            TimeValue(until_time=time(18, 0), value=1.0),
            TimeValue(until_time=END_OF_DAY, value=0.0),
        ]

        # Before first interval
        assert evaluate_time_values(time_values, time(6, 0), Interpolation.NO) == 0.0
        # During second interval
        assert evaluate_time_values(time_values, time(12, 0), Interpolation.NO) == 1.0
        # During third interval
        assert evaluate_time_values(time_values, time(20, 0), Interpolation.NO) == 0.0

    def test_at_exact_boundary(self) -> None:
        """Test evaluation at exact boundary transitions to next interval."""
        time_values = [
            TimeValue(until_time=time(8, 0), value=0.0),
            TimeValue(until_time=END_OF_DAY, value=1.0),
        ]

        # At exactly 8:00, should be in the second interval (value 1.0)
        assert evaluate_time_values(time_values, time(8, 0), Interpolation.NO) == 1.0

    def test_interpolation_midpoint(self) -> None:
        """Test linear interpolation at midpoint."""
        time_values = [
            TimeValue(until_time=time(12, 0), value=0.0),
            TimeValue(until_time=END_OF_DAY, value=1.0),
        ]

        # At 6:00 (halfway through first interval 0-12), interpolating from 0 to 0 = 0
        result = evaluate_time_values(time_values, time(6, 0), Interpolation.AVERAGE)
        assert result == 0.0  # prev_value=0.0, tv.value=0.0, so interpolated is 0.0

    def test_past_all_intervals_returns_last_value(self) -> None:
        """Test that time past all intervals returns last value."""
        time_values = [
            TimeValue(until_time=time(12, 0), value=0.5),
        ]

        # 13:00 is past the only interval
        result = evaluate_time_values(time_values, time(13, 0), Interpolation.NO)
        assert result == 0.5



"""Generate doc_locations.json from built idfkit-docs HTML.

Scans the idfkit-docs ``dist/`` directory for HTML anchor IDs that match
EnergyPlus object types and produces a mapping of
``{object_type: "relative/path/#anchor"}``.

The generated file is committed and shipped with the package so that
:mod:`idfkit.docs` can generate accurate documentation URLs without any
runtime dependency on idfkit-docs.

Usage::

    python -m idfkit.codegen.generate_doc_locations               # default: ../idfkit-docs/dist
    python -m idfkit.codegen.generate_doc_locations /path/to/dist  # explicit path
"""

from __future__ import annotations

import json
import re
import sys
from pathlib import Path

from idfkit.schema import get_schema
from idfkit.versions import LATEST_VERSION

# Regex to extract id="..." attributes from HTML.  We match
# lowercase-alphanumeric plus hyphens (the docs build convention for
# EnergyPlus object-type heading anchors).
_ANCHOR_RE = re.compile(r'id="([a-z0-9-]+)"')

# Same transform used by the docs build to turn object type names into
# heading anchors: strip colons, slashes, parens, and spaces.
_ANCHOR_STRIP_RE = re.compile(r"[:/() ]")

# Default location of idfkit-docs dist output relative to this repo.
_DEFAULT_DOCS_DIST = Path(__file__).resolve().parents[4] / "idfkit-docs" / "dist"

# Output path for the generated JSON.
_OUTPUT_PATH = Path(__file__).resolve().parents[1] / "schemas" / "doc_locations.json"


def _build_anchor_to_object_map() -> dict[str, str]:
    """Build a reverse mapping from anchor slug → canonical object type."""
    schema = get_schema(LATEST_VERSION)
    mapping: dict[str, str] = {}
    for obj_type in schema.object_types:
        anchor = _ANCHOR_STRIP_RE.sub("", obj_type.lower())
        mapping[anchor] = obj_type
    return mapping


def _find_latest_version_dir(dist_dir: Path) -> Path | None:
    """Find the latest versioned directory (e.g. ``v25.2``) in the dist."""
    version_dirs = sorted(
        (d for d in dist_dir.iterdir() if d.is_dir() and d.name.startswith("v")),
        key=lambda d: tuple(int(x) for x in d.name[1:].split(".") if x.isdigit()),
        reverse=True,
    )
    return version_dirs[0] if version_dirs else None


def generate(dist_dir: Path) -> dict[str, str]:
    """Scan *dist_dir* and return ``{object_type: relative_path}`` mapping."""
    version_dir = _find_latest_version_dir(dist_dir)
    if version_dir is None:
        print(f"Error: no versioned directories found in {dist_dir}", file=sys.stderr)
        sys.exit(1)

    io_ref = version_dir / "io-reference"
    if not io_ref.is_dir():
        print(f"Error: {io_ref} does not exist", file=sys.stderr)
        sys.exit(1)

    anchor_to_obj = _build_anchor_to_object_map()
    locations: dict[str, str] = {}

    # Walk all HTML files under io-reference/
    for html_path in sorted(io_ref.rglob("index.html")):
        content = html_path.read_text(encoding="utf-8", errors="replace")
        anchors = _ANCHOR_RE.findall(content)

        # Build the relative path from the versioned dir to this page
        page_dir = html_path.parent
        relative = page_dir.relative_to(version_dir)

        for anchor in anchors:
            obj_type = anchor_to_obj.get(anchor)
            if obj_type is not None and obj_type not in locations:
                locations[obj_type] = f"{relative}/#{anchor}"

    return dict(sorted(locations.items()))


def main() -> None:
    dist_dir = Path(sys.argv[1]) if len(sys.argv) > 1 else _DEFAULT_DOCS_DIST

    if not dist_dir.is_dir():
        print(f"Error: docs dist directory not found: {dist_dir}", file=sys.stderr)
        print("Build idfkit-docs first, or pass the path explicitly.", file=sys.stderr)
        sys.exit(1)

    locations = generate(dist_dir)
    print(f"Found {len(locations)} object type → docs page mappings")

    _OUTPUT_PATH.write_text(json.dumps(locations, separators=(",", ":"), sort_keys=True) + "\n", encoding="utf-8")
    print(f"Wrote {_OUTPUT_PATH}")


if __name__ == "__main__":
    main()



from __future__ import annotations


# --8<-- [start:example]
class ScheduleEvaluationError(Exception):
    """Raised when schedule cannot be evaluated."""

    pass


class UnsupportedScheduleType(ScheduleEvaluationError):
    """Schedule type not yet implemented."""

    pass


class ScheduleReferenceError(ScheduleEvaluationError):
    """Referenced schedule not found in document."""

    pass


class MalformedScheduleError(ScheduleEvaluationError):
    """Schedule syntax is invalid."""

    pass


# --8<-- [end:example]



"""SARIF 2.1.0 output formatter for compatibility lint diagnostics.

Produces `Static Analysis Results Interchange Format`__ (SARIF) output
that can be consumed by GitHub Code Scanning, VS Code SARIF Viewer,
and other SARIF-aware tools.

__ https://sarifweb.azurewebsites.net/
"""

from __future__ import annotations

import json
from typing import Any

from ._models import DIAGNOSTIC_CODES, CompatSeverity, Diagnostic

_SARIF_VERSION = "2.1.0"
_SARIF_SCHEMA = "https://raw.githubusercontent.com/oasis-tcs/sarif-spec/main/sarif-2.1/schema/sarif-schema-2.1.0.json"

_SEVERITY_MAP: dict[CompatSeverity, str] = {
    CompatSeverity.ERROR: "error",
    CompatSeverity.WARNING: "warning",
}


def _build_rules() -> list[dict[str, Any]]:
    """Build the ``rules`` array for the SARIF ``tool.driver`` block."""
    rules: list[dict[str, Any]] = []
    for code, description in DIAGNOSTIC_CODES.items():
        rules.append({
            "id": code,
            "shortDescription": {"text": description},
            "defaultConfiguration": {"level": "warning"},
        })
    return rules


def _diagnostic_to_result(diag: Diagnostic) -> dict[str, Any]:
    """Convert a single :class:`Diagnostic` to a SARIF ``result`` object."""
    result: dict[str, Any] = {
        "ruleId": diag.code,
        "level": _SEVERITY_MAP.get(diag.severity, "warning"),
        "message": {"text": diag.message},
        "locations": [
            {
                "physicalLocation": {
                    "artifactLocation": {"uri": diag.filename},
                    "region": {
                        "startLine": diag.line,
                        "startColumn": diag.col + 1,  # SARIF columns are 1-based
                        "endColumn": diag.end_col + 1,
                    },
                },
            },
        ],
        "properties": {
            "from_version": diag.from_version,
            "to_version": diag.to_version,
        },
    }
    if diag.suggested_fix is not None:
        result["fixes"] = [
            {
                "description": {"text": diag.suggested_fix},
            },
        ]
    return result


def format_sarif(diagnostics: list[Diagnostic], *, tool_version: str = "0.1.0") -> str:
    """Format a list of diagnostics as a SARIF 2.1.0 JSON string.

    Args:
        diagnostics: Lint diagnostics to format.
        tool_version: Version string for the tool metadata.

    Returns:
        A JSON string conforming to the SARIF 2.1.0 schema.
    """
    sarif: dict[str, Any] = {
        "$schema": _SARIF_SCHEMA,
        "version": _SARIF_VERSION,
        "runs": [
            {
                "tool": {
                    "driver": {
                        "name": "idfkit",
                        "informationUri": "https://github.com/idfkit/idfkit",
                        "version": tool_version,
                        "rules": _build_rules(),
                    },
                },
                "results": [_diagnostic_to_result(d) for d in diagnostics],
            },
        ],
    }
    return json.dumps(sarif, indent=2)



"""
Visualization utilities for EnergyPlus models.

This module provides SVG diagram generation for construction assemblies,
showing layer sequence, thicknesses, and thermal properties, as well as
interactive 3D building model visualization using plotly.

Example:
    >>> from idfkit import load_idf
    >>> from idfkit.visualization import construction_to_svg
    >>>
    >>> model = load_idf("building.idf")
    >>> wall = model["Construction"]["ExteriorWall"]
    >>>
    >>> svg = construction_to_svg(wall)
    >>> with open("wall_section.svg", "w") as f:
    ...     f.write(svg)

In Jupyter notebooks, constructions display automatically as SVG:
    >>> wall  # Displays graphical cross-section
"""

from __future__ import annotations

from .model import ColorBy, ModelViewConfig, view_exploded, view_floor_plan, view_model, view_normals
from .svg import SVGConfig, construction_to_svg, generate_construction_svg

__all__ = [
    "ColorBy",
    "ModelViewConfig",
    "SVGConfig",
    "construction_to_svg",
    "generate_construction_svg",
    "view_exploded",
    "view_floor_plan",
    "view_model",
    "view_normals",
]



from __future__ import annotations

from idfkit import IDFObject

zone: IDFObject = ...  # type: ignore[assignment]
# --8<-- [start:example]
print(zone.x_origin)
zone.x_origin = 5.0
# --8<-- [end:example]



from __future__ import annotations

from dataclasses import dataclass
from datetime import time
from idfkit import IDFObject


# --8<-- [start:example]
@dataclass
class CompactPeriod:
    """A 'Through:' block covering a date range."""

    end_month: int
    end_day: int
    day_rules: list[CompactDayRule]


@dataclass
class CompactDayRule:
    """A 'For:' block with day types and time-value pairs."""

    day_types: set[str]  # {"Weekdays", "Weekends", "Holidays", ...}
    time_values: list[tuple[time, float]]  # [(08:00, 0.0), (18:00, 1.0), ...]


def parse_compact(obj: IDFObject) -> list[CompactPeriod]:
    """Parse Schedule:Compact fields into structured data."""


# --8<-- [end:example]



"""
Reference graph for tracking object dependencies.

Provides O(1) lookups for:
- What objects reference a given name?
- What names does an object reference?
- Validation of reference integrity
"""

from __future__ import annotations

import logging
from collections import defaultdict
from collections.abc import Iterator
from typing import TYPE_CHECKING

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from .objects import IDFObject


class ReferenceGraph:
    """
    Tracks object references for instant dependency queries.

    The graph maintains two indexes:
    - _referenced_by: name -> set of objects that reference it
    - _references: object -> set of names it references

    This enables O(1) lookups for common operations like:
    - Finding all surfaces in a zone
    - Finding all objects using a construction
    - Detecting dangling references

    The reference graph is automatically maintained by
    [IDFDocument][idfkit.document.IDFDocument] when objects are added,
    removed, or when reference fields are modified.

    Examples:
        The reference graph automatically tracks which objects point
        to which names.  For instance, when a surface references a
        zone, that link is available for instant queries:

        >>> from idfkit import new_document
        >>> model = new_document()
        >>> model.add("Zone", "Perimeter_ZN_1")  # doctest: +ELLIPSIS
        Zone('Perimeter_ZN_1')
        >>> model.add("BuildingSurface:Detailed", "South_Wall",
        ...     surface_type="Wall", construction_name="",
        ...     zone_name="Perimeter_ZN_1",
        ...     outside_boundary_condition="Outdoors",
        ...     sun_exposure="SunExposed", wind_exposure="WindExposed",
        ...     validate=False)  # doctest: +ELLIPSIS
        BuildingSurface:Detailed('South_Wall')
        >>> model.references.is_referenced("Perimeter_ZN_1")
        True
        >>> stats = model.references.stats()
        >>> stats["total_references"] >= 1
        True
    """

    __slots__ = ("_object_lists", "_referenced_by", "_references")

    def __init__(self) -> None:
        # name (uppercase) -> set of (object, field_name) tuples that reference it
        self._referenced_by: dict[str, set[tuple[IDFObject, str]]] = defaultdict(set)
        # object -> set of (name_uppercase, field_name) tuples it references
        self._references: dict[IDFObject, set[tuple[str, str]]] = defaultdict(set)
        # object_list name -> set of object types that provide names for it
        self._object_lists: dict[str, set[str]] = defaultdict(set)

    def register_object_list(self, list_name: str, obj_type: str) -> None:
        """Register that an object type provides names for an object-list."""
        self._object_lists[list_name].add(obj_type)

    def register(self, obj: IDFObject, field_name: str, referenced_name: str) -> None:
        """
        Register that an object references another name.

        Args:
            obj: The object that contains the reference
            field_name: The field that contains the reference
            referenced_name: The name being referenced
        """
        if not referenced_name:
            return

        name_upper = referenced_name.upper()
        self._referenced_by[name_upper].add((obj, field_name))
        self._references[obj].add((name_upper, field_name))

    def unregister(self, obj: IDFObject) -> None:
        """Remove all reference tracking for an object."""
        if obj in self._references:
            # Remove from referenced_by
            for name_upper, field_name in self._references[obj]:
                if name_upper in self._referenced_by:
                    self._referenced_by[name_upper].discard((obj, field_name))
                    if not self._referenced_by[name_upper]:
                        del self._referenced_by[name_upper]
            del self._references[obj]

        # Also remove any references TO this object
        obj_name_upper = obj.name.upper() if obj.name else ""
        if obj_name_upper in self._referenced_by:
            del self._referenced_by[obj_name_upper]

    def get_referencing(self, name: str) -> set[IDFObject]:
        """
        O(1): Get all objects that reference a given name.

        Args:
            name: The name to look up

        Returns:
            Set of IDFObjects that reference this name

        Examples:
            Find all surfaces assigned to a zone (O(1)):

            >>> from idfkit import new_document
            >>> model = new_document()
            >>> model.add("Zone", "Perimeter_ZN_1")  # doctest: +ELLIPSIS
            Zone('Perimeter_ZN_1')
            >>> model.add("BuildingSurface:Detailed", "South_Wall",
            ...     surface_type="Wall", construction_name="",
            ...     zone_name="Perimeter_ZN_1",
            ...     outside_boundary_condition="Outdoors",
            ...     sun_exposure="SunExposed", wind_exposure="WindExposed",
            ...     validate=False)  # doctest: +ELLIPSIS
            BuildingSurface:Detailed('South_Wall')
            >>> len(model.references.get_referencing("Perimeter_ZN_1"))
            1
        """
        refs = self._referenced_by.get(name.upper(), set())
        return {obj for obj, _ in refs}

    def get_referencing_with_fields(self, name: str) -> set[tuple[IDFObject, str]]:
        """
        O(1): Get all (object, field_name) pairs that reference a given name.

        Args:
            name: The name to look up

        Returns:
            Set of (IDFObject, field_name) tuples
        """
        return self._referenced_by.get(name.upper(), set()).copy()

    def get_references(self, obj: IDFObject) -> set[str]:
        """
        O(1): Get all names that an object references.

        Args:
            obj: The object to look up

        Returns:
            Set of names (uppercase) that this object references
        """
        refs = self._references.get(obj, set())
        return {name for name, _ in refs}

    def get_references_with_fields(self, obj: IDFObject) -> set[tuple[str, str]]:
        """
        O(1): Get all (name, field_name) pairs that an object references.

        Args:
            obj: The object to look up

        Returns:
            Set of (name, field_name) tuples
        """
        return self._references.get(obj, set()).copy()

    def is_referenced(self, name: str) -> bool:
        """Check if a name is referenced by any object.

        Examples:
            Check whether a zone is used by any surface before deleting it:

            >>> from idfkit import new_document
            >>> model = new_document()
            >>> model.add("Zone", "Perimeter_ZN_1")  # doctest: +ELLIPSIS
            Zone('Perimeter_ZN_1')
            >>> model.references.is_referenced("Perimeter_ZN_1")
            False
            >>> model.add("BuildingSurface:Detailed", "South_Wall",
            ...     surface_type="Wall", construction_name="",
            ...     zone_name="Perimeter_ZN_1",
            ...     outside_boundary_condition="Outdoors",
            ...     sun_exposure="SunExposed", wind_exposure="WindExposed",
            ...     validate=False)  # doctest: +ELLIPSIS
            BuildingSurface:Detailed('South_Wall')
            >>> model.references.is_referenced("Perimeter_ZN_1")
            True
        """
        return name.upper() in self._referenced_by

    def get_dangling_references(self, valid_names: set[str]) -> Iterator[tuple[IDFObject, str, str]]:
        """
        Find all references to non-existent objects.

        Args:
            valid_names: Set of valid object names (uppercase)

        Yields:
            Tuples of (source_object, field_name, referenced_name)
        """
        valid_upper = {n.upper() for n in valid_names}

        for obj, refs in self._references.items():
            for name_upper, field_name in refs:
                if name_upper not in valid_upper:
                    yield (obj, field_name, name_upper)

    def rename_target(self, old_name: str, new_name: str) -> None:
        """
        Update indexes when a referenced target is renamed.

        Moves _referenced_by[OLD] -> _referenced_by[NEW] and updates
        corresponding _references entries for all affected objects.

        Args:
            old_name: The old target name
            new_name: The new target name
        """
        old_upper = old_name.upper()
        new_upper = new_name.upper()
        if old_upper == new_upper:
            return

        referrers = self._referenced_by.pop(old_upper, set())
        if not referrers:
            return

        # Update _references for each referring object
        for obj, field_name in referrers:
            obj_refs = self._references.get(obj)
            if obj_refs is not None:
                obj_refs.discard((old_upper, field_name))
                obj_refs.add((new_upper, field_name))

        # Merge into new key (there may already be refs to new_name)
        if new_upper in self._referenced_by:
            self._referenced_by[new_upper].update(referrers)
        else:
            self._referenced_by[new_upper] = referrers

    def update_reference(self, obj: IDFObject, field_name: str, old_value: str | None, new_value: str | None) -> None:
        """
        Update indexes when an object's reference field changes.

        Removes the old entry from both indexes and adds the new entry.

        Args:
            obj: The object whose field changed
            field_name: The field that changed
            old_value: The previous referenced name (or None)
            new_value: The new referenced name (or None)
        """
        # Remove old
        if old_value:
            old_upper = old_value.upper()
            refs_set = self._referenced_by.get(old_upper)
            if refs_set is not None:
                refs_set.discard((obj, field_name))
                if not refs_set:
                    del self._referenced_by[old_upper]
            obj_refs = self._references.get(obj)
            if obj_refs is not None:
                obj_refs.discard((old_upper, field_name))

        # Add new
        if new_value and new_value.strip():
            new_upper = new_value.upper()
            self._referenced_by[new_upper].add((obj, field_name))
            self._references[obj].add((new_upper, field_name))

    def clear(self) -> None:
        """Clear all reference tracking."""
        self._referenced_by.clear()
        self._references.clear()
        self._object_lists.clear()

    def __len__(self) -> int:
        """Return total number of references tracked."""
        return sum(len(refs) for refs in self._references.values())

    def stats(self) -> dict[str, int]:
        """Return statistics about the reference graph."""
        return {
            "total_references": len(self),
            "objects_with_references": len(self._references),
            "names_referenced": len(self._referenced_by),
            "object_lists": len(self._object_lists),
        }



from __future__ import annotations

from idfkit.simulation import EnergyPlusConfig

config: EnergyPlusConfig = ...  # type: ignore[assignment]
# --8<-- [start:example]
from idfkit.simulation import find_energyplus

config = find_energyplus()
print(f"Version: {config.version}")
print(f"Path: {config.executable}")
# --8<-- [end:example]



from __future__ import annotations

from idfkit import IDFDocument

model: IDFDocument = ...  # type: ignore[assignment]
# --8<-- [start:example]
import logging
from idfkit.simulation import simulate, SimulationProgress

logger = logging.getLogger("simulation")


def on_progress(event: SimulationProgress) -> None:
    logger.info(
        "simulation_progress",
        extra={
            "phase": event.phase,
            "percent": event.percent,
            "environment": event.environment,
            "message": event.message,
        },
    )


result = simulate(model, "weather.epw", on_progress=on_progress)
# --8<-- [end:example]
